"""
navdoon.collector
-----------------
Define collectors, that collect Statsd requests and queue them to be
processed by the processor.
"""

import os
import socket
from socket import socket as SocketClass
from abc import abstractmethod, ABCMeta
from threading import Event
from navdoon.pystdlib.queue import Queue
from navdoon.utils.common import LoggerMixIn
from navdoon.utils.system import ExpandableThreadPool
from navdoon.pystdlib.typing import Dict, Any, Tuple, List, Optional

DEFAULT_PORT = 8125


def socket_type_repr(socket_type):
    # type: (int) -> str
    sock_types = {
        socket.SOCK_STREAM: "TCP",
        socket.SOCK_DGRAM: "UDP"
    }
    return sock_types.get(socket_type, "UNKNOWN")


class AbstractCollector(object):
    """Abstract base class for collectors"""

    __metaclass__ = ABCMeta

    def __init__(self):
        self._queue = Queue()  # type: Queue

    @abstractmethod
    def start(self):
        # type: () -> None
        raise NotImplementedError

    @abstractmethod
    def wait_until_queuing_requests(self, timeout=None):
        # type: (float) -> None
        raise NotImplementedError

    @abstractmethod
    def shutdown(self):
        # type: () -> None
        raise NotImplementedError

    @abstractmethod
    def wait_until_shutdown(self, timeout=None):
        # type: (float) -> None
        raise NotImplementedError

    @property
    def queue(self):
        # type: () -> Queue
        return self._queue

    @queue.setter
    def queue(self, value):
        # type: (Queue) -> None
        for method in ('put_nowait',):
            if not callable(getattr(value, method, None)):
                raise ValueError(
                    "Invalid queue for collector. Queue is missing "
                    "method '{}'".format(method))
        self._queue = value

    def __repr__(self):
        return "collector <{}>".format(self.__class__)


class SocketServer(LoggerMixIn, AbstractCollector):
    """Collect Statsd metrics via TCP/UDP socket"""

    def __init__(self, **kargs):
        # type: (**Dict[str, Any]) -> None
        AbstractCollector.__init__(self)
        LoggerMixIn.__init__(self)
        self.chunk_size = 8196  # type: int
        self.socket_type = socket.SOCK_DGRAM  # type: int
        self.socket_timeout = 1  # type: float
        self.host = '127.0.0.1'  # type: str
        self.port = DEFAULT_PORT  # type: int
        self.user = None  # type: int
        self.group = None  # type: int
        self.socket = None  # type: socket.socket
        self._stop_queuing_requests = Event()  # type: Event
        self._queuing_requests = Event()  # type: Event
        self._shutdown = Event()  # type: Event
        self._should_shutdown = Event()  # type: Event
        self.configure(**kargs)  # type: Dict[str, Any]
        self.log_signature = "collector.socket_server "  # type: str
        self.num_worker_threads = 4  # type: int
        self.worker_threads_limit = 128  # type: int

    def __del__(self):
        # type: () -> None
        self._do_shutdown()

    def __repr__(self):
        # type: () -> str
        return "collector.socket_server {}@{}:{}".format(
            socket_type_repr(self.socket_type),
            self.host,
            self.port
        )

    def configure(self, **kargs):
        # type: (**Dict[str, Any]) -> List[str]
        """Configure the server, setting attributes.
        Returns a list of attribute names that were affected
        """
        configured = []
        for key in ('host', 'port', 'user', 'group', 'socket_type',
                    'num_worker_threads', 'worker_threads_limit'):
            if key in kargs:
                setattr(self, key, kargs[key])
                configured.append(key)
        return configured

    def start(self):
        # type: () -> None
        self._bind_socket()
        try:
            while not self._should_shutdown.is_set():
                self._shutdown.clear()
                self._log("starting serving requests on {} {}:{}".format(
                    socket_type_repr(self.socket_type), self.host, self.port))
                self._pre_start()
                if self.socket_type == socket.SOCK_STREAM:
                    self._queue_requests_tcp()
                else:
                    self._queue_requests_udp()
                self._post_start()
                self._log("stopped serving requests")
        finally:
            self._do_shutdown()

    def is_queuing_requests(self):
        # type: () -> bool
        return self._queuing_requests.is_set()

    def wait_until_queuing_requests(self, timeout=None):
        # type: (float) -> None
        self._queuing_requests.wait(timeout)

    def shutdown(self):
        # type: () -> None
        self._log_debug("shutting down ...")
        self._stop_queuing_requests.set()
        self._should_shutdown.set()

    def wait_until_shutdown(self, timeout=None):
        # type: (float) -> None
        self._log_debug("waiting until shutdown ...")
        self._shutdown.wait(timeout)
        self._log("shutdown successfully")

    def _pre_start(self):
        # type: () -> None
        pass

    def _queue_requests_udp(self):
        # type: () -> None
        should_stop = self._stop_queuing_requests.is_set
        chunk_size = self.chunk_size
        receive = self.socket.recv
        timeout_exception = socket.timeout
        enqueue = self._queue.put_nowait

        try:
            self._queuing_requests.set()
            self._log_debug("starting queuing UDP requests ...")
            while not should_stop():
                try:
                    data = receive(chunk_size)
                except timeout_exception:
                    data = None
                if data:
                    enqueue(data.decode())
        finally:
            self._log_debug("stopped queuing UDP requests")
            self._queuing_requests.clear()

    def _queue_requests_tcp(self):
        # type: () -> None
        stop_event = self._stop_queuing_requests
        should_stop_accepting = stop_event.is_set
        chunk_size = self.chunk_size
        queue_put_nowait = self._queue.put_nowait
        shutdown_rdwr = socket.SHUT_RDWR
        socket_timeout_exception = socket.timeout

        thread_pool = ExpandableThreadPool(self.num_worker_threads)
        thread_pool.workers_limit = self.worker_threads_limit
        thread_pool.logger = self.logger
        thread_pool.log_signature = "threadpool =< {} ".format(self)
        thread_pool.initialize()

        def _enqueue_from_connection(conn, address):
            # type: (socket.socket, Tuple[str, int]) -> None
            buffer_size = chunk_size
            enqueue = queue_put_nowait
            timeout_exception = socket_timeout_exception
            should_stop_queuing = stop_event.is_set
            receive = conn.recv
            incomplete_line_chunk = u''
            try:
                self._log_debug("collecting metrics from TCP {}:{} ...".format(address[0], address[1]))
                while not should_stop_queuing():
                    try:
                        buff_bytes = receive(buffer_size)
                    except timeout_exception:
                        continue
                    if not buff_bytes:
                        break

                    buff_lines = buff_bytes.decode().splitlines(True)
                    if incomplete_line_chunk != '':
                        buff_lines[0] = incomplete_line_chunk + buff_lines[0]
                        incomplete_line_chunk = ''

                    if not buff_lines[-1].endswith('\n'):
                        incomplete_line_chunk = buff_lines.pop()

                    enqueue(''.join(buff_lines))
            finally:
                conn.shutdown(shutdown_rdwr)
                conn.close()
                if incomplete_line_chunk != '':
                    enqueue(incomplete_line_chunk)

        try:
            self._queuing_requests.set()
            self._log_debug("starting accepting TCP connections ...")
            while not should_stop_accepting():
                try:
                    (connection, remote_addr) = self.socket.accept()
                    self._log_debug("TCP connection from {}:{} ...".format(remote_addr[0], remote_addr[1]))
                    connection.settimeout(self.socket_timeout)
                except socket_timeout_exception:
                    continue
                thread_pool.do(_enqueue_from_connection, connection, remote_addr)
            self._log_debug("stopped accepting TCP connection")
            thread_pool.stop(True, 10)  # TODO: set this timeout from object attrs
            self._log_debug("stopped enqueuing TCP requests")
        finally:
            self._queuing_requests.clear()

    def _post_start(self):
        # type: () -> None
        pass

    def _do_shutdown(self):
        # type: () -> None
        self._close_socket()
        self._shutdown.set()

    def _bind_socket(self, sock=None):
        # type: (Optional[SocketClass]) -> None
        if not sock:
            sock = self._create_socket()
        self._close_socket()
        self._change_process_user_group()
        self.socket = sock
        self._log("bound to address {}:{}".format(*sock.getsockname()))

    def _create_socket(self):
        # type: () -> SocketClass
        sock = socket.socket(socket.AF_INET, self.socket_type)
        sock.bind((self.host, self.port))
        sock.settimeout(self.socket_timeout)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        if self.socket_type == socket.SOCK_STREAM:
            sock.listen(5)
        return sock

    def _close_socket(self):
        # type: () -> None
        sock = self.socket
        if sock:
            try:
                sock.shutdown(socket.SHUT_RDWR)
            except socket.error as e:
                pass
            finally:
                sock.close()
        self.socket = None

    def _change_process_user_group(self):
        # type: () -> None
        if self.user:
            self._log("changing process user to {}".format(self.user))
            os.seteuid(self.user)
        if self.group:
            self._log("changing process group to {}".format(self.group))
            os.setegid(self.group)



"""
navdoon.server
--------------
Define the Statsd server, that uses other components (collector, processor,
destination) to handle Statsd requests and flush metrics to specified
destinations.
"""

import multiprocessing
from time import time, sleep
from threading import Thread, RLock, Event
from navdoon.pystdlib import queue
from navdoon.collector import AbstractCollector
from navdoon.utils.common import LoggerMixIn
from navdoon.processor import QueueProcessor
from navdoon.pystdlib.typing import List, Optional, Union
from navdoon.pystdlib.queue import Queue


def validate_collectors(collectors):
    """Validate collectors to be usable by the server"""
    # type: (List[Any]) -> None
    for collector in collectors:
        if not isinstance(collector, AbstractCollector):
            raise ValueError(
                "Collectors should extend AbstractCollector")


class Server(LoggerMixIn):
    """Statsd server
    Accepts multiple collectors, and a queue processor instance, creates a queue and shares it amongst all
    collectors and the processor.
    """

    def __init__(self):
        # type: () -> None
        super(Server, self).__init__()
        self.log_signature = 'server '  # type: str
        self._collectors = []  # type: List[AbstractCollector]
        self._running_collectors = []  # type: List[AbstractCollector]
        self._running = Event()  # type: Event
        self._shutdown = Event()  # type: Event
        self._reload = Event()  # type: Event
        self._pause = Event()  # type: Event
        self._should_reload = Event()  # type: Event
        self._running_lock = RLock()  # type: RLock
        self._pause_lock = RLock()  # type: RLock
        self._queue = self._create_queue()  # type: Queue
        self._queue_processor = None  # type: QueueProcessor
        self._running_queue_processor = None  # type: QueueProcessor

    @property
    def queue_processor(self):
        # type: () -> QueueProcessor
        return self._queue_processor

    @queue_processor.setter
    def queue_processor(self, value):
        # type: (QueueProcessor) -> None
        if not isinstance(value, QueueProcessor):
            raise ValueError(
                "Invalid queue value. Processor should extend QueueProcessor")
        self._queue_processor = value

    def set_collectors(self, collectors):
        # type: (List[AbstractCollector]) -> Server
        validate_collectors(collectors)
        self._collectors = collectors
        return self

    def start(self):
        # type: () -> None
        """Start the queue processor in background then starts all the collectors.
        Blocks current thread, until shutdown() is called from another thread.
        """
        if not self._collectors:
            raise Exception("Can not start Statsd server without a collector")
        if self._queue_processor is None:
            self._log_warn(
                "no queue processor provided. Creating a queue processor "
                "with no destinations")
            self._queue_processor = self.create_queue_processor()
        reloading = False
        keep_running = True
        with self._running_lock:
            self._log("starting ...")
            while keep_running:
                self._should_reload.clear()
                self._share_queue()
                queue_thread = self._start_queue_processor()
                self._log_debug("started queue processor thread")
                try:
                    collector_threads = self._start_collectors()
                    self._running.set()
                    if reloading:
                        self._reload.set()
                    self._log("collectors are running")
                    self._pause.wait()
                    self._pause.clear()
                    self._shutdown_collectors()
                    for thread in collector_threads:
                        thread.join()
                    self._running_collectors = []
                    self._shutdown_queue_processor()
                finally:
                    self._log("stopped, joining queue processor thread")
                    queue_thread.join()
                    self._running.clear()
                reloading = self._should_reload.is_set()
                keep_running = reloading
                if reloading:
                    self._log("reloading ...")

    def _start_collectors(self):
        # type: () -> List[Thread]
        """Start collectors in background threads and returns the threads"""
        collector_threads = []
        self._log_debug("starting {} collectors ...".format(len(self._collectors)))
        for collector in self._collectors:
            thread = Thread(target=collector.start)
            collector_threads.append(thread)
            thread.start()
            collector.wait_until_queuing_requests()
            self._running_collectors.append(collector)
        return collector_threads

    def is_running(self):
        # type: () -> bool
        return self._running.is_set()

    def wait_until_running(self, timeout=None):
        # type: (Optional[float]) -> None
        self._running.wait(timeout)

    def shutdown(self, timeout=None):
        # type: (Optional[float]) -> None
        """Shutdown the server, stopping the collectors and the queue processor"""
        with self._pause_lock:
            self._log("shutting down ...")
            start_time = time()
            self._shutdown_collectors(timeout)
            if self._queue_processor.is_processing():
                queue_timeout = max(0.1, timeout - (time() - start_time)) if timeout else None
                self._shutdown_queue_processor(queue_timeout)
            else:
                self._log_debug("queue processor is not processing")
            self._pause.set()
            self._close_queue()
            self._shutdown.set()

    def wait_until_shutdown(self, timeout=None):
        # type: (Optional[float]) -> None
        start = time()
        self._shutdown.wait(timeout)
        while True:
            if not self._running.is_set():
                break
            if timeout is not None and time() - start > timeout:
                raise Exception("Server shutdown timeout")
            sleep(0.5)

    def create_queue_processor(self):
        # type: () -> QueueProcessor
        processor = QueueProcessor(self._queue)
        processor.logger = self.logger
        return processor

    def reload(self):
        # type: () -> None
        with self._pause_lock:
            self._reload.clear()
            self._should_reload.set()
            self._pause.set()

    def wait_until_reload(self, timeout=None):
        # type: (Optional[float]) -> None
        self._reload.wait(timeout)

    @staticmethod
    def _use_multiprocessing():
        # type: () -> bool
        # FIXME: use multiprocessing if available
        # return available_cpus() > 1
        return False

    @classmethod
    def _create_queue(cls):
        # type: () -> Queue
        return multiprocessing.Queue() if cls._use_multiprocessing() else queue.Queue()

    def _share_queue(self):
        # type: () -> None
        queue_ = self._queue
        self._queue_processor.queue = queue_
        for collector in self._collectors:
            collector.queue = queue_

    def _close_queue(self):
        # type: () -> None
        queue_ = self._queue
        if queue_:
            if callable(getattr(queue_, 'close', None)):
                queue_.close()  # type: ignore
        self._queue = None

    def _start_queue_processor(self):
        # type: () -> Union[multiprocessing.Process, Thread]
        if self._use_multiprocessing():
            self._log_debug("starting queue processor in a separate process")
            queue_process = multiprocessing.Process(target=self._queue_processor.process)
            queue_process.start()
        else:
            self._log_debug("starting queue processor in a thread")
            queue_process = Thread(target=self._queue_processor.process)  # type: ignore
            queue_process.start()

        self._queue_processor.wait_until_processing(30)
        if not self._queue_processor.is_processing():
            self._log_error("queue processor didn't start correctly time is up")
            self._queue_processor.shutdown()
            queue_process.join()
            raise Exception("Failed to start the queue processor")
        self._running_queue_processor = self._queue_processor
        return queue_process

    def _shutdown_queue_processor(self, timeout=None):
        # type: (Optional[float]) -> bool
        if not self._running_queue_processor:
            return False
        start_time = time()
        self._log_debug("shutting down the queue processor ...")
        self._running_queue_processor.shutdown()
        self._running_queue_processor.wait_until_shutdown(timeout)
        if self._running_queue_processor.is_processing():
            self._log_error(
                "Queue processor shutdown timeout after {} seconds".format(
                    time() - start_time))
            raise Exception(
                "Server shutdown timeout when shutting down processor")
        self._running_queue_processor = None
        return True

    def _shutdown_collectors(self, timeout=None):
        # type: (Optional[float]) -> None
        if not self._running_collectors:
            return

        start_time = time()
        self._log_debug("shutting down {} collectors ...".format(len(
            self._running_collectors)))

        stopped_collectors = []
        try:
            for collector in self._running_collectors:
                self._log_debug("shutting down {}".format(collector))
                collector.shutdown()
                collector.wait_until_shutdown(timeout)
                self._log("{} shutdown successfully!".format(collector))
                stopped_collectors.append(collector)
                if timeout is None:
                    continue
                time_elapsed = time() - start_time
                if time_elapsed > timeout:
                    self._log_error(
                        "collectors shutdown timeout after "
                        "{} seconds".format(time_elapsed))
                    raise Exception(
                        "Server shutdown timed out when "
                        "shutting down collectors")
                else:
                    timeout -= time_elapsed
            self._log_debug("all collectors shutdown")
        finally:
            for collector in stopped_collectors:
                self._running_collectors.remove(collector)



#!/usr/bin/env python
from __future__ import absolute_import, print_function

import sys
import os
import signal
import subprocess
import gc
from os.path import dirname, abspath
from random import randint
from tempfile import mkstemp
from time import time, sleep
from re import match
from unittest import TestCase, main, skip

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from statsdmetrics.client import Client
from statsdmetrics.client.tcp import TCPClient

TEST_DIR = dirname(abspath(__file__))
BASE_DIR = dirname(TEST_DIR)
CONF_FILE = os.path.join(TEST_DIR, 'fixtures', 'functional_test_config.ini')
APP_BIN = os.path.join(BASE_DIR, 'bin', 'navdoon_src')


def metrics_to_dict(metrics):
    result = dict()
    for metric in metrics:
        name, value, _ = metric.split()
        result[name] = float(value) if '.' in value else int(value)
    return result


def wait_until_log_matches(process, pattern, timeout=10, sleep_time=None):
    start_time = time()
    logs = []
    while True:
        if time() - start_time > timeout:
            logs = [log for log in logs if log.strip()]
            print("server logs dump")
            print("================")
            for log in logs:
                print(log, file=sys.stderr)
            raise RuntimeError(
                "waiting for pattern {} in server logs timedout.".format(
                    pattern))
        line = process.stderr.readline().strip()
        logs.append(line)
        if match(pattern, line):
            break
        if sleep_time:
            sleep(sleep_time)


def wait_until_server_shuts_down(process, timeout=20):
    wait_until_log_matches(process, '.*server shutdown successfully', timeout, 0.05)


def wait_until_server_starts_collecting(process, timeout=10):
    wait_until_log_matches(process, '.*server.+collectors.+are\s+running', timeout, 0.05)


def wait_until_server_processed_metric(process, name, timeout=10):
    wait_until_log_matches(process, '.*queue.processor processing metrics:\s*' + name, timeout, 0.1)


def write_to_file(filename, content):
    with open(filename, 'wt') as fh:
        fh.write(content)


class TestNavdoonStatsdServer(TestCase):
    @staticmethod
    def create_server_program_args(config=CONF_FILE, udp_port=None, tcp_port=None, flush_interval=None, flush_file=None, flush_file_csv=None):
        program_args = [APP_BIN]
        if flush_interval is not None:
            program_args.extend(('--flush-interval', str(flush_interval)))
        if flush_file is not None:
            program_args.extend(('--flush-file', str(flush_file)))
        if flush_file_csv is not None:
            program_args.extend(('--flush-file-csv', str(flush_file_csv)))
        if config is not None:
            program_args.extend(('--config', config))
        if udp_port is not None:
            program_args.extend(('--collect-udp', '127.0.0.1:{}'.format(udp_port)))
        if tcp_port is not None:
            program_args.extend(('--collect-tcp', '127.0.0.1:{}'.format(tcp_port)))
        return program_args

    def create_server_process(self, *args, **kwargs):
        app_args = self.create_server_program_args(*args, **kwargs)
        app_process = subprocess.Popen(app_args, stderr=subprocess.PIPE, stdout=subprocess.PIPE,
                                       universal_newlines=True)
        try:
            wait_until_server_starts_collecting(app_process)
        except RuntimeError as error:
            app_process.kill()
            raise error
        return app_process

    def setUp(self):
        self.app_process = None
        self.remove_files = []

    def tearDown(self):
        if self.app_process is not None:
            self.app_process.poll()
            if self.app_process.returncode is None:
                self.app_process.kill()
            self.app_process = None
            gc.collect()
        for filename in self.remove_files:
            if os.path.exists(filename):
                os.unlink(filename)

    def test_udp_collectors_flushing_stdout(self):
        udp_port = randint(8125, 8999)
        flush_interval = 2
        self.app_process = self.create_server_process(udp_port=udp_port, flush_interval=flush_interval)

        client = Client("localhost", udp_port)
        for _ in range(0, 3):
            client.increment("event")
        client.timing("process", 101)
        client.timing("process", 102)
        client.timing("process", 103)
        # wait for at least 1 flush
        sleep(flush_interval)
        self.app_process.terminate()
        wait_until_server_shuts_down(self.app_process)
        flushed_metrics = self.app_process.communicate()[0].splitlines()
        self.assertGreater(len(flushed_metrics), 5, 'flushed 1 counter and at least 4 timers')
        self.assertDictEqual(
            {
                "event": 3, "process.count": 3, "process.max": 103,
                "process.min": 101, "process.mean": 102,
                "process.median": 102
            },
            metrics_to_dict(flushed_metrics)
        )

    def test_udp_and_tcp_collectors_combine_and_flush_to_stdout(self):
        udp_port = randint(8125, 8999)
        tcp_port = randint(8125, 9999)
        flush_interval = 2
        self.app_process = self.create_server_process(udp_port=udp_port, tcp_port=tcp_port,
                                                      flush_interval=flush_interval)
        client = Client("localhost", udp_port)
        tcp_client = TCPClient("localhost", tcp_port)
        for _ in range(0, 2):
            client.increment("event")
            tcp_client.increment("event")

        client.timing("process", 85)
        client.timing("process", 98)
        tcp_client.timing("process", 87)
        tcp_client.timing("query", 2)
        # wait for at least 1 flush
        sleep(flush_interval)
        self.app_process.terminate()
        wait_until_server_shuts_down(self.app_process)
        flushed_metrics = self.app_process.communicate()[0].splitlines()
        self.assertGreater(len(flushed_metrics), 9, 'flushed 1 counter and at least 2 x 4 timers')

        self.maxDiff = None
        self.assertDictEqual(
            {
                "event": 4, "process.count": 3, "process.max": 98,
                "process.min": 85, "process.mean": 90,
                "process.median": 87, "query.count": 1, "query.max": 2.0,
                "query.min": 2.0, "query.mean": 2.0, "query.median": 2.0
            },
            metrics_to_dict(flushed_metrics)
        )

    def test_reload_server_keeps_the_queue(self):
        tcp_port = randint(8125, 9999)
        _, config_filename = mkstemp(text=True)
        self.remove_files.append(config_filename)
        write_to_file(
            config_filename,
            """
[navdoon]
log-level=DEBUG
log-stderr=true
flush-stdout=true
collect-tcp=127.0.0.1:{}
flush-interval=60
""".format(tcp_port)
        )
        self.app_process = self.create_server_process(config=config_filename)
        tcp_client = TCPClient("localhost", tcp_port)
        for _ in range(0, 2):
            tcp_client.increment("event")
        tcp_client.timing("query", 2)
        del tcp_client
        wait_until_server_processed_metric(self.app_process, 'event')

        udp_port = randint(8125, 8999)
        flush_interval = 5
        write_to_file(
            config_filename,
            """
[navdoon]
log-level=DEBUG
log-stderr=true
flush-stdout=true
collect-udp=127.0.0.1:{}
flush-interval={}
""".format(udp_port, flush_interval)
        )

        self.app_process.send_signal(signal.SIGHUP)
        wait_until_server_starts_collecting(self.app_process, 10)

        client = Client("localhost", udp_port)
        for _ in range(0, 2):
            client.increment("event")
        client.timing("query", 4)
        client.increment("finish")

        self.assertRaises(Exception, TCPClient, "localhost", tcp_port)  # TCP collector should be down
        os.remove(config_filename)

        wait_until_server_processed_metric(self.app_process, 'finish')
        # wait for at least 1 flush
        sleep(flush_interval)

        self.app_process.terminate()
        wait_until_server_shuts_down(self.app_process)

        flushed_metrics = self.app_process.communicate()[0].splitlines()
        self.assertGreaterEqual(len(flushed_metrics), 5)

        self.maxDiff = None
        self.assertDictEqual(
            {
                "event": 4, "query.count": 2, "query.max": 4.0,
                "query.min": 2.0, "query.mean": 3.0, "query.median": 3.0,
                "finish": 1
            },
            metrics_to_dict(flushed_metrics)
        )

    def test_flushing_files(self):
        _, file_name = mkstemp()
        os.remove(file_name)
        udp_port = randint(8125, 8999)
        flush_interval = 2
        csv_file_name = file_name + '.csv'
        self.remove_files.append(file_name)
        self.remove_files.append(csv_file_name)
        self.app_process = self.create_server_process(udp_port=udp_port, flush_interval=flush_interval, flush_file=file_name, flush_file_csv=csv_file_name)

        client = Client("localhost", udp_port)
        for _ in range(0, 3):
            client.increment("event")
        client.timing("process", 20)
        client.timing("process", 22)
        client.timing("process", 24)
        # wait for at least 1 flush
        sleep(flush_interval)
        self.app_process.terminate()
        wait_until_server_shuts_down(self.app_process)

        self.assertTrue(os.path.exists(file_name))
        with open(file_name) as file_handle:
            flushed_metrics = [line.rstrip() for line in file_handle.readlines()]

        self.assertTrue(os.path.exists(csv_file_name))
        with open(csv_file_name) as file_handle:
            flushed_metrics_csv = [line.rstrip() for line in file_handle.readlines()]

        expected_metrics_dict = {
                "event": 3, "process.count": 3, "process.max": 24,
                "process.min": 20, "process.mean": 22,
                "process.median": 22
            }
        self.assertGreater(len(flushed_metrics), 5, 'flushed 1 counter and at least 4 timers')
        self.assertDictEqual(expected_metrics_dict,  metrics_to_dict(flushed_metrics))

        self.assertGreater(len(flushed_metrics_csv), 5, 'flushed 1 counter and at least 4 timers')


if __name__ == '__main__':
    main()



import sys
import unittest
import logging
from random import random
from time import sleep, time
from threading import Event, Thread

from statsdmetrics import Counter
from navdoon.destination import AbstractDestination
from navdoon.processor import QueueProcessor
from navdoon.collector import AbstractCollector
from navdoon.server import Server, validate_collectors


class StubDestination(AbstractDestination):
    """Stub destination to inspect server behavior"""

    def __init__(self, expected_count=0):
        self.metrics = []
        self.expected_count = expected_count
        self._flushed_expected_count = Event()

    def flush(self, metrics):
        self.metrics.extend(metrics)
        if len(self.metrics) >= self.expected_count:
            self._flushed_expected_count.set()

    def wait_until_expected_count_items(self, timeout=None):
        self._flushed_expected_count.wait(timeout)


class StubCollector(AbstractCollector):
    """Stub collector to inspect server behavior"""

    def __init__(self, data=None, frequency=1.0):
        super(StubCollector, self).__init__()
        self.data = data
        self.max_items = None
        self.frequency = float(frequency)
        self._shutdown = Event()
        self._running = Event()
        self.queued_data = []

    def start(self):
        self._running.set()
        while not self._shutdown.is_set():
            if self.max_items is None or self.max_items > len(
                    self.queued_data):
                self.queue.put(self.data)
                self.queued_data.append(self.data)
            sleep(random() / self.frequency)
        self._running.clear()

    def wait_until_queuing_requests(self, timeout=None):
        pass

    def shutdown(self):
        self._shutdown.set()

    def is_queuing_requests(self):
        return self._running.is_set()

    def wait_until_shutdown(self, timeout=None):
        self._shutdown.wait(timeout)


class TestFunctions(unittest.TestCase):
    """Test server module functions"""

    def test_validate_collectors_fails_on_non_collectors(self):
        self.assertRaises(ValueError, validate_collectors, ["not a collector"])
        self.assertRaises(ValueError, validate_collectors, [2])

    def test_validate_collectors(self):
        collector = StubCollector()
        validate_collectors([collector])


class TestServer(unittest.TestCase):
    """Test Server"""

    def test_create_queue_processor(self):
        server = Server()
        server.logger = logging.getLogger('test')
        processor = server.create_queue_processor()
        self.assertIsInstance(processor, QueueProcessor)
        self.assertEqual(server.logger, processor.logger)

    def test_start_fails_without_collectors(self):
        server = Server()
        processor = server.create_queue_processor()
        processor.set_destinations([StubDestination()])
        server.queue_processor = processor
        self.assertRaises(Exception, server.start)

    def test_server_creates_a_default_queue_processor_without_destination(
            self):
        server = Server()
        collector = StubCollector('-')
        server.set_collectors([collector])
        server_thread = Thread(target=server.start)
        server_thread.setDaemon(True)
        server_thread.start()
        server.wait_until_running(5)
        self.assertTrue(server.is_running())
        server.shutdown()
        server.wait_until_shutdown(5)
        self.assertFalse(server.is_running())
        self.assertFalse(server_thread.isAlive())

    def test_start_and_shutdown_with_configured_queue_processor(self):
        start_time = time()
        destination = StubDestination(10)
        server = Server()
        processor = server.create_queue_processor()
        processor.set_destinations([destination])
        processor.flush_interval = 0.5
        server.queue_processor = processor
        metric = Counter('test.metric', 1)
        collector = StubCollector(data=metric.to_request(), frequency=10)
        server.set_collectors([collector])
        server_thread = Thread(target=server.start)
        server_thread.setDaemon(True)
        server_thread.start()
        server.wait_until_running(5)
        self.assertTrue(server.is_running())
        destination.wait_until_expected_count_items(10)

        server.shutdown()
        server.wait_until_shutdown(5)
        self.assertFalse(server.is_running())
        self.assertFalse(server_thread.isAlive())
        self.assertFalse(collector.is_queuing_requests())
        self.assertFalse(processor.is_processing())

        self.assertGreaterEqual(len(destination.metrics), 10)
        for metric in destination.metrics:
            self.assertEqual(3, len(metric))
            self.assertEqual(metric[0], "test.metric")
            self.assertGreaterEqual(metric[1], 0)
            self.assertGreaterEqual(metric[2], start_time)

    def test_reload_stops_collectors_and_processor_and_new_ones_used_in_server_start_thread(self):
        start_time = time()
        destination = StubDestination(10)
        server = Server()
        processor = server.create_queue_processor()
        processor.set_destinations([destination])
        processor.flush_interval = 0.5
        server.queue_processor = processor
        metric = Counter('test.metric', 1)
        collector = StubCollector(data=metric.to_request(), frequency=10)
        server.set_collectors([collector])
        server_thread = Thread(target=server.start)
        server_thread.setDaemon(True)
        server_thread.start()
        server.wait_until_running(5)
        self.assertTrue(server.is_running())
        destination.wait_until_expected_count_items(10)

        metric2 = Counter('test.metric2', 1)
        collector2 = StubCollector(data=metric2.to_request(), frequency=10)
        destination2 = StubDestination(5)
        processor2 = server.create_queue_processor()
        processor2.set_destinations([destination2])
        processor2.flush_interval = 0.5
        server.queue_processor = processor2
        server.set_collectors([collector2])
        server.reload()
        server.wait_until_reload(10)
        self.assertTrue(server_thread.isAlive())
        self.assertTrue(server.is_running())
        self.assertTrue(collector2.is_queuing_requests())
        self.assertTrue(processor2.is_processing())
        self.assertFalse(collector.is_queuing_requests())
        self.assertFalse(processor.is_processing())
        destination2.wait_until_expected_count_items(15)

        self.assertEqual(len(destination2.metrics), 5)
        for metric in destination2.metrics:
            self.assertEqual(3, len(metric))
            self.assertEqual(metric[0], "test.metric2")
            self.assertGreaterEqual(metric[1], 0)
            self.assertGreaterEqual(metric[2], start_time)



import sys
import socket
import threading
import unittest
import logging
import gc
from navdoon.pystdlib.queue import Empty, Queue
from navdoon.collector import SocketServer


def find_open_port(host, sock_type):
    for port in range(8125, 65535):
        try:
            sock = socket.socket(socket.AF_INET, sock_type)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
            sock.bind((host, port))
            break
        except socket.error:
            pass
        finally:
            sock.close()
    return port


def consume_queue(queue_, count, timeout=1):
    consumed = []
    for i in range(count):
        try:
            consumed.append(queue_.get(True, timeout))
        except Empty:
            break
    return tuple(consumed)


def socket_sendall_close(sock, data_set):
    for data in data_set:
        sock.sendall(data)
    sock.shutdown(socket.SHUT_RDWR)
    sock.close()


def create_logger():
    logger = logging.Logger('navdoon.tests')
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stderr))
    return logger


class SocketServerTestCaseMixIn(object):
    def setup_socket_server(self, socket_type):
        self.host = '127.0.0.1'
        self.port = find_open_port(self.host, socket_type)
        self.server = SocketServer()
        self.server.socket_type = socket_type
        self.server.host = self.host
        self.server.port = self.port
        self.server_thread = threading.Thread(target=self.server.start)
        self.server_thread.daemon = True

    def start_server_send_data_and_consume_queue(self, data_set, socket_type):
        self.server_thread.start()
        self.server.wait_until_queuing_requests()
        client_sock = socket.socket(socket.AF_INET, socket_type)
        client_sock.connect((self.host, self.port))
        socket_sendall_close(client_sock, data_set)
        in_queue = consume_queue(self.server.queue, len(data_set))
        return in_queue


class TestUDPServer(SocketServerTestCaseMixIn, unittest.TestCase):
    def setUp(self):
        self.setup_socket_server(socket.SOCK_DGRAM)

    def tearDown(self):
        if self.server.is_queuing_requests():
            self.server.shutdown()
            self.server.wait_until_shutdown(3)
        self.server = None
        gc.collect()

    def test_constructor_args(self):
        conf = dict(user='thisuser',
                    port=9876,
                    group='thatgroup',
                    host='example.org')
        server = SocketServer(**conf)
        self.assertEqual(server.host, 'example.org')
        self.assertEqual(server.port, 9876)
        self.assertEqual(server.user, 'thisuser')
        self.assertEqual(server.group, 'thatgroup')

    def test_configure(self):
        conf = dict(user='someuser',
                    port=1234,
                    group='somegroup',
                    host='example.org')
        configured = self.server.configure(**conf)
        self.assertEqual(sorted(configured), sorted(conf.keys()))
        self.assertEqual(self.server.host, 'example.org')
        self.assertEqual(self.server.port, 1234)
        self.assertEqual(self.server.user, 'someuser')
        self.assertEqual(self.server.group, 'somegroup')

    def test_get_set_queue(self):
        def set_queue(queue_):
            self.server.queue = queue_

        self.assertRaises(ValueError, set_queue, "not a queue")
        queue = Queue()
        self.server.queue = queue
        self.assertEqual(queue, self.server.queue)

    def test_queue_requests(self):
        data_set = ("test message".encode(), "could be anything".encode())
        expected_values_in_queue = tuple([data.decode() for data in data_set])
        in_queue = self.start_server_send_data_and_consume_queue(
            data_set, socket.SOCK_DGRAM)
        self.assertEqual(expected_values_in_queue, in_queue)


class TestTCPServer(SocketServerTestCaseMixIn, unittest.TestCase):
    def setUp(self):
        self.setup_socket_server(socket.SOCK_STREAM)

    def tearDown(self):
        if self.server.is_queuing_requests():
            self.server.shutdown()
            self.server.wait_until_shutdown(3)
        self.server = None
        gc.collect()

    def test_constructor_args(self):
        conf = dict(socket_type=socket.SOCK_STREAM)
        server = SocketServer(**conf)
        self.assertEqual(server.socket_type, socket.SOCK_STREAM)

    def test_configure(self):
        conf = dict(socket_type=socket.SOCK_STREAM)
        configured = self.server.configure(**conf)
        self.assertEqual(sorted(configured), sorted(conf.keys()))
        self.assertEqual(self.server.socket_type, socket.SOCK_STREAM)

    def test_queue_requests(self):
        data_set = ("test message\nin 2 lines\n".encode(), "resource.cpu 42|g\n".encode())
        expected_values_in_queue = ''.join([data.decode() for data in data_set])
        in_queue = self.start_server_send_data_and_consume_queue(
            data_set, socket.SOCK_STREAM)
        self.assertEqual(expected_values_in_queue, ''.join(in_queue))

    def test_queue_requests_includes_incomplete_lines(self):
        long_request = "test message with long message\n" * 500
        # the long request will break into 2 items in the queue, since the test needs to know the number of items in the
        # queue to pick, I'm putting an extra item in the data_set so while test consumes the queue, it will get 1 extra
        # item from the queue to match the number of items put on the queue. looks ugly but I'm fine with it now
        data_set = (long_request.encode(), "resource.cpu 42|g".encode(), "".encode())
        expected_values_in_queue = ''.join([data.decode() for data in data_set])
        in_queue = self.start_server_send_data_and_consume_queue(
            data_set, socket.SOCK_STREAM)
        self.assertEqual(expected_values_in_queue, ''.join(in_queue))

    def test_shutdown(self):
        data_set = ("".encode(), "test_messsage".encode(), "name 10|c@0.1".encode())
        expected_values_in_queue = ''.join([data.decode() for data in data_set])
        in_queue = self.start_server_send_data_and_consume_queue(
            data_set, socket.SOCK_STREAM)
        self.assertEqual(expected_values_in_queue, ''.join(in_queue))
        self.server.shutdown()
        self.server.wait_until_shutdown(5)
        self.assertFalse(self.server.is_queuing_requests())



import unittest
import navdoon.utils.common


class TestDataSeries(unittest.TestCase):
    def test_mean(self):
        simple = navdoon.utils.common.DataSeries([10,20,30])
        self.assertEqual(20, simple.mean())

        negative = navdoon.utils.common.DataSeries([-2,0.8,12, 100])
        self.assertEqual(27.7, negative.mean())

        single = navdoon.utils.common.DataSeries([18.4])
        self.assertEqual(18.4, single.mean())

    def test_max(self):
        simple = navdoon.utils.common.DataSeries([10,20,30])
        self.assertEqual(30, simple.max())

        negative = navdoon.utils.common.DataSeries([-2,0.6,0, 13.2])
        self.assertEqual(13.2, negative.max())

        single = navdoon.utils.common.DataSeries([17.9])
        self.assertEqual(17.9, single.max())

    def test_min(self):
        simple = navdoon.utils.common.DataSeries([10,20,30])
        self.assertEqual(10, simple.min())

        negative = navdoon.utils.common.DataSeries([-2,0.6,0, 13.2])
        self.assertEqual(-2, negative.min())

        single = navdoon.utils.common.DataSeries([14.6])
        self.assertEqual(14.6, single.min())

    def test_median(self):
        simple = navdoon.utils.common.DataSeries([10,20,30])
        self.assertEqual(20, simple.median())

        with_float = navdoon.utils.common.DataSeries([2,0.6,0, 13.2])
        self.assertEqual(7.6, with_float.median())

        single = navdoon.utils.common.DataSeries([12.8])
        self.assertEqual(12.8, single.median())

        double = navdoon.utils.common.DataSeries([12.8, 14])
        self.assertEqual(13.4, double.median())



"""
navdoon.destination
-------------------
Where the data received from Statsd will be
flushed to
"""
from navdoon.destination.abstract import AbstractDestination
from navdoon.destination.graphite import Graphite
from navdoon.destination.stream import Stream, Stdout, CsvStream, CsvStdout
from navdoon.destination.file import TextFile, CsvFile

__all__ = ['Graphite', 'Stream', 'Stdout', 'CsvStream', 'CsvStdout', 'TextFile', 'CsvFile']



"""
navdoon.utils.common
--------------------
Common utilities and mixin classes
"""

import socket
from os import getpid
from time import sleep
from logging import INFO, DEBUG, ERROR, WARN
from threading import Lock
from navdoon.pystdlib.typing import AnyStr, Sequence


class LoggerMixIn(object):
    """A MixIn class for anything that needs to log messages"""

    def __init__(self):
        self.logger = None  # type: logging.Logger
        self.log_pattern = "{signature}{message}"  # type: str
        self.log_signature = ''  # type: str
        self._pid = getpid()  # type: int

    def _log_debug(self, msg):
        # type: (str) -> None
        self._log(msg, DEBUG)

    def _log_error(self, msg):
        # type: (str) -> None
        self._log(msg, ERROR)

    def _log_warn(self, msg):
        # type: (str) -> None
        self._log(msg, WARN)

    def _log(self, msg, level=INFO):
        # type: (str, int) -> None
        if self.logger:
            self.logger.log(
                level,
                self.log_pattern.format(message=msg,
                                        signature=self.log_signature,
                                        pid=self._pid))


class TCPClient(LoggerMixIn):
    """A generic TCP client with reconnecting feature"""

    def __init__(self, host, port):
        # type: (str, int) -> None
        super(TCPClient, self).__init__()
        self.host = host  # type: str
        self.port = port  # type: int
        self.max_retry = None  # type: int
        self._connection_tries = 0  # type: int
        self._sleep_between_retries = 0.5  # type: float
        self._sock = None  # type: socket.socket
        self._connection_lock = Lock()  # type: Lock
        self._sending_lock = Lock()  # type: Lock

    def connect(self):
        self._connection()
        self._log("connected to {}:{}".format(self.host, self.port))

    def disconnect(self):
        with self._connection_lock:
            self._log_debug("disconnecting from {}:{}".format(self.host,
                                                              self.port))
            if self._sock:
                try:
                    self._sock.shutdown(socket.SHUT_RDWR)
                    self._sock.close()
                except socket.error:
                    pass
            self._sock = None
            self._log("disconnected from {}:{}".format(self.host, self.port))

    def reconnect(self):
        self._log_debug("reconnecting to {}:{}".format(self.host, self.port))
        if self._connection_tries >= self.max_retry:
            raise IOError(
                "Reached maximum connection tries of '{}' to {}:{}".format(
                    self.max_retry, self.host, self.port))
        self.disconnect()
        self.connect()

    def reset_connection_tries(self):
        with self._connection_lock:
            self._connection_tries = 0

    def _send_with_lock(self, data_bytes):
        # type: (bytes) -> None
        data_size = len(data_bytes)
        with self._sending_lock:
            while True:
                self._log_debug("sending {} bytes to {}:{} ...".format(
                    data_size, self.host, self.port))
                sock = self._connection()
                try:
                    sock.sendall(data_bytes)
                    self._log_debug("sent {} bytes to {}:{}".format(
                        data_size, self.host, self.port))
                    break
                except socket.error as err:
                    self._log_error("failed to send data to {}:{}. {}".format(
                        self.host, self.port, err))
                    self.reconnect()

    def _connection(self):
        # type: () -> socket.socket
        with self._connection_lock:
            if not self._sock:
                max_retry = self.max_retry
                while True:
                    self._connection_tries += 1
                    self._log_debug("connecting to {}:{} try {}/{}".format(
                        self.host, self.port, self._connection_tries,
                        max_retry))
                    try:
                        sock = socket.socket(socket.AF_INET,
                                             socket.SOCK_STREAM)
                        sock.connect((self.host, self.port))
                        self._log_debug("connected to {}:{}".format(self.host,
                                                                    self.port))
                        self._sock = sock
                        break
                    except socket.error as err:
                        self._log_error(
                            "failed to connect to {}:{}. {}".format(
                                self.host, self.port, err))
                        sock.close()
                        if max_retry and max_retry <= self._connection_tries:
                            raise IOError(
                                "Reached maximum connection tries of "
                                "'{}' to {}:{}".format(
                                    max_retry, self.host, self.port))
                    sleep(self._sleep_between_retries * self._connection_tries)
        return self._sock


class DataSeries(object):
    def __init__(self, data):
        # type: (Sequence[float]) -> None
        self._count = len(data)  # type: int
        if self._count < 1:
            raise ValueError("Can not create a series from an empty data set")
        self._data = sorted(data)  # type: Sequence[float]

    def count(self):
        # type: () -> int
        return self._count

    def min(self):
        # type: () -> float
        return min(self._data)

    def max(self):
        # type: () -> float
        return max(self._data)

    def mean(self):
        # type: () -> float
        return sum(self._data) / self._count

    def median(self):
        # type: () -> float
        count = self._count
        if count == 2:
            return self.mean()
        last_index = count - 1
        middle_index = count // 2
        if middle_index < last_index and count % 2 == 0:
            return (self._data[middle_index] + self._data[middle_index + 1]) / 2
        else:
            return self._data[middle_index]



import unittest
import logging
import sys
from time import time, sleep
from threading import Thread, Event
from statsdmetrics import Counter, Set, Gauge, GaugeDelta, Timer
from navdoon.pystdlib.queue import Queue
from navdoon.processor import QueueProcessor, StatsShelf
from navdoon.utils.common import LoggerMixIn
from navdoon.destination import AbstractDestination


def create_debug_logger():
    logger = logging.Logger('navdoon.test')
    logger.addHandler(logging.StreamHandler(sys.stderr))
    logger.setLevel(logging.DEBUG)
    return logger


class DestinationWithoutFlushMethod(object):
    pass


class StubDestination(LoggerMixIn, AbstractDestination):
    def __init__(self, expected_count=0):
        super(StubDestination, self).__init__()
        self.log_signature = 'test.destination '
        self.metrics = []
        self.expected_count = expected_count
        self.flushed_expected_count = Event()

    def flush(self, metrics):
        self._log_debug("received {} metrics".format(len(metrics)))
        self.metrics.extend(metrics)
        if len(self.metrics) >= self.expected_count:
            self.flushed_expected_count.set()
        else:
            self.flushed_expected_count.clear()

    def wait_until_expected_count_items(self, timeout=None):
        self._log(
            "flush destination waiting for expected items to be flushed ...")
        self.flushed_expected_count.wait(timeout)


class TestQueueProcessor(unittest.TestCase):
    """Test processor.QueueProcessor class"""

    def test_set_flush_interval_accepts_positive_numbers(self):
        processor = QueueProcessor(Queue())
        processor.flush_interval = 103
        self.assertEqual(103, processor.flush_interval)
        processor.flush_interval = 0.58
        self.assertEqual(0.58, processor.flush_interval)
        processor.flush_interval = '3.4'
        self.assertEqual(3.4, processor.flush_interval)

    def test_set_flush_interval_fails_on_not_positive_numbers(self):
        processor = QueueProcessor(Queue())

        def set_interval(value):
            processor.flush_interval = value

        self.assertRaises(ValueError, set_interval, 0)
        self.assertRaises(ValueError, set_interval, -10)
        self.assertRaises(ValueError, set_interval, "not a number")

    def test_processor_does_not_accept_invalid_queue(self):
        self.assertRaises(ValueError, QueueProcessor, "not a queue")
        self.assertRaises(ValueError, QueueProcessor, 100)

        processor = QueueProcessor(Queue())

        def set_queue(value):
            processor.queue = value

        self.assertRaises(ValueError, set_queue, "not a queue")
        self.assertRaises(ValueError, set_queue, 100)

    def test_set_the_queue(self):
        queue_ = Queue()
        processor = QueueProcessor(queue_)
        self.assertEqual(queue_, processor.queue)

        next_queue = Queue()
        processor.queue = next_queue
        self.assertEqual(next_queue, processor.queue)

    def test_set_destination_fails_on_invalid_destination(self):
        processor = QueueProcessor(Queue())
        self.assertRaises(ValueError, processor.set_destinations,
                          "not a destination")

    def test_set_destinations_when_not_started(self):
        processor = QueueProcessor(Queue())
        destinations = [StubDestination()]
        processor.set_destinations(destinations)
        self.assertEqual(destinations, processor._destinations)
        self.assertFalse(processor.is_processing())

    def test_destinations_can_change_when_queue_processor_is_running(self):
        processor = QueueProcessor(Queue())
        try:
            destinations = [StubDestination()]
            processor.set_destinations(destinations)
            processor_thread = Thread(target=processor.process)
            processor_thread.start()
            processor.wait_until_processing(10)
            self.assertTrue(processor.is_processing())
            new_destinations = [StubDestination()]
            processor.set_destinations(new_destinations)
            self.assertEqual(new_destinations, processor.get_destinations())
            self.assertTrue(processor.is_processing())
        finally:
            processor.shutdown()

    def test_clear_destinations(self):
        destination = StubDestination()
        queue_ = Queue()
        processor = QueueProcessor(queue_)
        processor.set_destinations([destination])
        self.assertEqual([destination], processor._destinations)
        processor.clear_destinations()
        self.assertEqual([], processor._destinations)

    def test_process(self):
        expected_flushed_metrics_count = 2
        metrics = (Counter('user.jump', 2),
                   Set('username', 'navdoon'),
                   Set('username', 'navdoon.test'),
                   Counter('user.jump', 4),
                   Set('username', 'navdoon'),
                   Counter('user.jump', -1),)
        queue_ = Queue()
        destination = StubDestination()
        destination.expected_count = expected_flushed_metrics_count
        processor = QueueProcessor(queue_)
        processor.set_destinations([destination])
        processor.init_destinations()
        process_thread = Thread(target=processor.process)
        process_thread.start()
        processor.wait_until_processing(5)
        for metric in metrics:
            queue_.put(metric.to_request())
        destination.wait_until_expected_count_items(5)
        processor.shutdown()
        processor.wait_until_shutdown(5)
        self.assertEqual(expected_flushed_metrics_count,
                         len(destination.metrics))
        self.assertEqual(('user.jump', 5), destination.metrics[0][:2])
        self.assertEqual(('username', 2), destination.metrics[1][:2])

    def test_process_stops_on_stop_token_in_queue(self):
        token = 'STOP'
        expected_flushed_metrics_count = 2
        metrics = (Counter('user.login', 1),
                   Set('username', 'navdoon'),
                   Counter('user.login', 3),
                   token,
                   Counter('user.login', -1),
                   Counter('user.logout', 1),)
        queue_ = Queue()
        destination = StubDestination()
        destination.expected_count = expected_flushed_metrics_count
        processor = QueueProcessor(queue_)
        processor.flush_interval = 2
        processor.stop_process_token = token
        processor.set_destinations([destination])
        process_thread = Thread(target=processor.process)
        process_thread.start()
        processor.wait_until_processing(5)
        for metric in metrics:
            if metric is token:
                sleep(processor.flush_interval)  # make sure one flush happened before token
                request = metric
            else:
                request = metric.to_request()
            queue_.put(request)
        # make sure the processor has process the queue
        processor.wait_until_shutdown(5)
        self.assertFalse(processor.is_processing())
        destination.wait_until_expected_count_items(10)
        self.assertEqual(expected_flushed_metrics_count,
                         len(destination.metrics))
        self.assertEqual(('user.login', 4), destination.metrics[0][:2])
        self.assertEqual(('username', 1), destination.metrics[1][:2])

    def test_continues_processing_after_reload(self):
        metrics = (Counter('user.login', 1), Set('username', 'navdoon'),
                   Counter('user.login', 3))
        queue_ = Queue()
        destination = StubDestination()
        destination.expected_count = 1

        processor = QueueProcessor(queue_)
        processor.flush_interval = 1
        processor.set_destinations([destination])
        process_thread = Thread(target=processor.process)
        process_thread.start()
        processor.wait_until_processing(5)

        expected_flushed_metrics2 = 2
        destination2 = StubDestination()
        destination2.expected_count = expected_flushed_metrics2
        processor.set_destinations([destination2])

        for metric in metrics:
            queue_.put(metric.to_request())

        destination.wait_until_expected_count_items(5)
        processor.shutdown()
        process_thread.join(5)
        self.assertFalse(processor.is_processing())
        self.assertGreaterEqual(len(destination.metrics), 1)
        self.assertLessEqual(queue_.qsize(), 2)

        for metric in metrics:
            queue_.put(metric.to_request())

        self.assertGreaterEqual(queue_.qsize(), len(metrics))
        resume_process_thread = Thread(target=processor.process)
        resume_process_thread.start()
        processor.wait_until_processing(5)
        self.assertTrue(processor.is_processing())
        self.assertEqual(processor.get_destinations(), [destination2])

        destination2.wait_until_expected_count_items(5)
        processor.shutdown()
        resume_process_thread.join(5)
        self.assertGreaterEqual(len(destination2.metrics), expected_flushed_metrics2)

    def test_queue_can_only_change_when_not_processing(self):
        processor = QueueProcessor(Queue())
        orig_queue = processor.queue
        process_thread = Thread(target=processor.process)
        process_thread.start()
        processor.wait_until_processing(5)

        def set_queue(new_queue):
            processor.queue = new_queue

        new_queue = Queue()
        self.assertRaises(Exception, set_queue, new_queue)
        self.assertEqual(orig_queue, processor.queue)

        processor.shutdown()
        processor.wait_until_shutdown(5)

        processor.queue = new_queue
        self.assertEqual(new_queue, processor.queue)

    def test_process_timers(self):
        start_timestamp = time()
        expected_flushed_metrics_count = 2 + 5  # each timer has 5 separate metrics
        metrics = (Counter('user.jump', 2),
                   Set('username', 'navdoon'),
                   Timer('db.query', 300),
                   Set('username', 'navdoon2'),
                   Counter('user.jump', -1),
                   Timer('db.query', 309),
                   Timer('db.query', 303)
                   )
        queue_ = Queue()
        destination = StubDestination()
        destination.expected_count = expected_flushed_metrics_count
        processor = QueueProcessor(queue_)
        processor.set_destinations([destination])
        process_thread = Thread(target=processor.process)
        process_thread.start()
        processor.wait_until_processing(5)
        for metric in metrics:
            queue_.put(metric.to_request())
        destination.wait_until_expected_count_items(5)
        processor.shutdown()
        processor.wait_until_shutdown(5)
        self.assertEqual(expected_flushed_metrics_count,
                         len(destination.metrics))

        metrics_dict = dict()
        for (name, value, timestamp) in destination.metrics:
            metrics_dict[name] = value
            self.assertGreaterEqual(timestamp, start_timestamp)

        self.assertEqual(metrics_dict['user.jump'], 1)
        self.assertEqual(metrics_dict['username'], 2)
        self.assertEqual(metrics_dict['db.query.count'], 3)
        self.assertEqual(metrics_dict['db.query.max'], 309)
        self.assertEqual(metrics_dict['db.query.min'], 300)
        self.assertEqual(metrics_dict['db.query.mean'], 304)
        self.assertEqual(metrics_dict['db.query.median'], 303)


class TestStatsShelf(unittest.TestCase):
    def test_counters(self):
        shelf = StatsShelf()
        self.assertEqual(dict(), shelf.counters())

        shelf.add(Counter("mymetric", 3))
        shelf.add(Counter("mymetric", 2))
        shelf.add(Counter("something.else", 2, 0.5))
        expected = {"mymetric": 5, "something.else": 4}
        self.assertEqual(expected, shelf.counters())

        counters = shelf.counters()
        counters["counters should"] = "not changed"
        self.assertEqual(expected, shelf.counters())

    def test_sets(self):
        shelf = StatsShelf()
        self.assertEqual(dict(), shelf.sets())

        shelf.add(Set("users", "me"))
        shelf.add(Set("users", "me"))
        shelf.add(Set("users", "you"))
        shelf.add(Set("say.what?", "nothing"))
        shelf.add(Set("users", "me"))
        shelf.add(Set("say.what?", "nothing"))
        shelf.add(Set("say.what?", "ok"))
        expected = {"users": {"me", "you"},
                    "say.what?": {"nothing", "ok"}}
        self.assertEqual(expected, shelf.sets())

        sets = shelf.sets()
        sets["sets should"] = set("not change")
        self.assertEqual(expected, shelf.sets())

    def test_gauges(self):
        shelf = StatsShelf()
        self.assertEqual(dict(), shelf.gauges())

        shelf.add(Gauge("cpu%", 50))
        shelf.add(Gauge("cpu%", 51))
        shelf.add(Gauge("mem%", 20))
        shelf.add(Gauge("mem%", 23))
        shelf.add(Gauge("cpu%", 58))

        expected = {"cpu%": 58, "mem%": 23}
        self.assertEqual(expected, shelf.gauges())

        gauges = shelf.gauges()
        gauges["gauges should"] = "not change"
        self.assertEqual(expected, shelf.gauges())

    def test_gauge_deltas(self):
        shelf = StatsShelf()
        self.assertEqual(dict(), shelf.gauges())

        shelf.add(GaugeDelta("cpu%", 10))
        shelf.add(Gauge("mem%", 10))
        shelf.add(GaugeDelta("cpu%", 10))
        shelf.add(GaugeDelta("cpu%", -5))
        shelf.add(GaugeDelta("mem%", -2))
        shelf.add(GaugeDelta("mem%", 4))

        expected = {"cpu%": 15, "mem%": 12}
        self.assertEqual(expected, shelf.gauges())

    def test_timers_data(self):
        shelf = StatsShelf()
        self.assertEqual(dict(), shelf.timers_data())

        shelf.add(Timer("query.user", 3.223))
        shelf.add(Timer("query.user", 4.12))
        shelf.add(Timer("api.auth", 9.78))
        shelf.add(Timer("api.auth", 8.45))

        expected = {"query.user": [3.223, 4.12], "api.auth": [9.78, 8.45]}
        self.assertEqual(expected, shelf.timers_data())

    def test_timers(self):
        shelf = StatsShelf()
        self.assertEqual(dict(), shelf.timers())

        shelf.add(Timer("query.user", 2.9))
        shelf.add(Timer("query.user", 3))
        shelf.add(Timer("query.user", 4.1))
        shelf.add(Timer("api.auth", 8))
        shelf.add(Timer("api.auth", 7.96))
        shelf.add(Timer("cache.clear", 1.56))

        expected = {
            "query.user": dict(count=3, min=2.9, max=4.1, mean=3.3333333333333335, median=3),
            "api.auth": dict(count=2, min=7.96, max=8, mean=7.98, median=7.98),
            "cache.clear": dict(count=1, min=1.56, max=1.56, mean=1.56, median=1.56),
        }
        self.assertEqual(expected, shelf.timers())

    def test_clear_all_metrics(self):
        shelf = StatsShelf()

        shelf.add(Set("users", "me"))
        shelf.add(Counter("mymetric", 3))
        shelf.add(Timer("query", 4.12))
        shelf.add(Gauge("cpu%", 38))

        shelf.clear()

        self.assertEqual(dict(), shelf.counters())
        self.assertEqual(dict(), shelf.sets())
        self.assertEqual(dict(), shelf.timers_data())
        self.assertEqual(dict(), shelf.gauges())


if __name__ == '__main__':
    unittest.main()


"""
navdoon.app
-----------
Define the application, what is executed and handles interations and signals.
The application combines all the other components into what the end user
can use as the final product.
"""

from __future__ import print_function

import socket
import sys
import logging
import logging.handlers
import signal
from argparse import ArgumentParser, FileType, Namespace
from threading import RLock, Thread
from time import sleep

import navdoon
from navdoon.pystdlib import configparser
from navdoon.server import Server
from navdoon.destination import Stdout, Graphite, AbstractDestination, TextFile, CsvFile
from navdoon.collector import AbstractCollector, SocketServer, DEFAULT_PORT
from navdoon.utils.common import LoggerMixIn
from navdoon.utils.system import os_syslog_socket
from navdoon.pystdlib.typing import Tuple, IO, Dict, Any, Set

LOG_LEVEL_NAMES = ('DEBUG', 'INFO', 'WARN', 'ERROR', 'FATAL', 'CRITICAL')  # type: Tuple[str, str, str, str, str, str]


def parse_config_file(file_):
    # type: (IO) -> Dict[str, Any]
    """Parse app configurations from contents of a file"""

    parser = configparser.ConfigParser()  # type: ignore
    _parse_method = getattr(parser, 'read_file', parser.readfp)
    _parse_method(file_)

    config = dict()
    for key, value in parser.items('navdoon'):
        if key == 'config':
            continue
        elif key == 'log-level' and value not in LOG_LEVEL_NAMES:
            raise ValueError(
                "Invalid log level '{}' in configuration file '{}'".format(
                    value, file_.name))
        config[key.replace('-', '_')] = value
    return config


def default_syslog_socket():
    # type: () -> str
    return os_syslog_socket() or 'localhost:{}'.format(
        logging.handlers.SYSLOG_UDP_PORT)  # type: ignore


class App(LoggerMixIn):
    """Navdoon application, configure and use the components based on the
    provided arguments and configurations.
    """

    def __init__(self, args):
        # type: (List[str]) -> None
        LoggerMixIn.__init__(self)
        self.log_signature = "navdoon.app "  # type: str
        self._config = dict()  # type: Dict[str, Any]
        self._args = args  # type: List[str]
        self._server = None  # type: Server
        self._run_lock = RLock()  # type: RLock
        self._shutdown_lock = RLock()  # type: RLock
        self._reload_lock = RLock()  # type: RLock
        self._configure(args)

    def __del__(self):
        # type() -> None
        self.shutdown(1)
        self._close_logger()

    @staticmethod
    def get_description():
        # type: () -> str
        return "{} v{}\n{}".format(navdoon.__title__, navdoon.__version__,
                                   navdoon.__summary__)

    @staticmethod
    def get_default_config():
        # type: () -> Dict[str, Any]
        return dict(config=None,
                    log_level='INFO',
                    log_file=None,
                    log_stderr=False,
                    log_syslog=False,
                    syslog_socket=default_syslog_socket(),
                    flush_interval=1,
                    flush_stdout=False,
                    flush_graphite='',
                    flush_file='',
                    flush_file_csv='',
                    collect_udp='',
                    collect_tcp='',
                    collector_threads=4,
                    collector_threads_limit=128)

    def get_args(self):
        # type: () -> List[Any]
        return self._args

    def get_config(self):
        # type: () -> Dict[str, Any]
        return self._config

    def get_logger(self):
        # type: () -> logging.Logger
        if not self.logger:
            self.logger = self._create_logger()  # type: logging.Logger
        return self.logger

    def create_destinations(self):
        # type: () -> List[AbstractDestination]
        destinations = []  # type: List[AbstractDestination]
        if self._config.get('flush_stdout'):
            destinations.append(Stdout())
        if self._config.get('flush_graphite'):
            for graphite_address in self._config['flush_graphite'].split(','):
                graphite_address = graphite_address.strip().split(':')
                graphite_host = graphite_address.pop(0).strip()
                graphite_port = graphite_address and int(graphite_address.pop(
                )) or 2003
                destinations.append(Graphite(graphite_host, graphite_port))
        if self._config.get('flush_file'):
            for file_path in self._config['flush_file'].split('|'):
                destinations.append(TextFile(file_path))
        if self._config.get('flush_file_csv'):
            for file_path in self._config['flush_file_csv'].split('|'):
                destinations.append(CsvFile(file_path))
        return destinations

    def create_collectors(self):
        # type: () -> List[AbstractCollector]
        collectors = []  # type: List[AbstractCollector]
        if self._config.get('collect_tcp'):
            tcp_collectors = self._create_socket_servers(
                self._config['collect_tcp'], socket.SOCK_STREAM)
            for collector in tcp_collectors:
                collector.num_worker_threads = self._config['collector_threads']
                collector.worker_threads_limit = self._config['collector_threads_limit']
            collectors.extend(tcp_collectors)
        if self._config.get('collect_udp'):
            collectors.extend(
                self._create_socket_servers(
                    self._config['collect_udp'], socket.SOCK_DGRAM))
        if len(collectors) < 1:
            collectors.extend(
                self._create_socket_servers(
                    '127.0.0.1:8125', socket.SOCK_DGRAM))
        return collectors

    def _create_socket_servers(self, addresses, socket_type):
        # type: (str, int) -> List[SocketServer]
        collectors = []
        socket_addresses = self.get_addresses_with_unique_ports(
            addresses)
        for (host, port) in socket_addresses:
            socket_server = self._configure_socket_server(
                SocketServer(),
                host=host,
                port=port
            )
            socket_server.socket_type = socket_type
            collectors.append(socket_server)
        return collectors

    def create_server(self):
        # type: () -> Server
        server = Server()
        return self._configure_server(server)

    def run(self):
        # type: () -> int
        with self._run_lock:
            self._register_signal_handlers()
            self._server = self.create_server()
            server_thread = Thread(target=self._server.start)
            server_thread.start()
            # If we block by joining thread, we won't receive OS signals
            # poll the status of the server thread every often instead.
            while server_thread.is_alive():
                sleep(2)
            server_thread.join()
        return 0

    def shutdown(self, timeout=None):
        # type: (float) -> None
        if not self._server:
            return
        with self._shutdown_lock:
            self._log_debug("shutting down server ...")
            self._server.shutdown(timeout)
            self._server = None
            self._log("server shutdown successfully")

    def reload(self):
        # type: () -> None
        with self._reload_lock:
            if not self._server:
                raise Exception("App is not running, can not reload")
            self._configure(self._args)
            logger = self._create_logger()
            self._close_logger()
            self.logger = logger
            self._configure_server(self._server)
            self._server.reload()

    def _configure_server(self, server):
        # type: (Server) -> Server
        conf = self.get_config()
        logger = self.get_logger()
        destinations = self.create_destinations()
        if logger.handlers:
            server.logger = logger
        queue_processor = server.queue_processor
        if queue_processor is None:
            queue_processor = server.create_queue_processor()
        queue_processor.logger = logger
        queue_processor.flush_interval = conf['flush_interval']
        queue_processor.set_destinations(destinations)
        server.queue_processor = queue_processor
        server.set_collectors(self.create_collectors())
        return server

    def _configure_socket_server(self, collector, host=None, port=None):
        # type: (SocketServer, str, int) -> SocketServer
        logger = self.get_logger()
        if logger.handlers:
            collector.logger = logger
        if host is not None:
            collector.host = host
        if port is not None:
            collector.port = port
        return collector

    def _configure(self, args):
        # type: (List[str]) -> None
        parsed_args = vars(self._parse_args(args))
        configs = self.get_default_config()
        if parsed_args['config'] is not None:
            with parsed_args['config'] as config_file:
                configs.update(parse_config_file(config_file))

        store_true_args = ('log_stderr', 'log_syslog', 'flush_stdout')
        for key, value in parsed_args.items():
            if key in store_true_args:
                if key not in configs or value is True:
                    configs[key] = value
            elif value is not None:
                configs[key] = value

        self._validate_configs(configs)
        self._config = configs

    def _parse_args(self, args):
        # type: (List[str]) -> Namespace
        parser = ArgumentParser(description=self.get_description())
        parser.add_argument('-c',
                            '--config',
                            help='path to config file',
                            type=FileType('rt'))
        parser.add_argument('--log-level',
                            help='logging level',
                            choices=LOG_LEVEL_NAMES)
        parser.add_argument('--log-file', help='path to log file')
        parser.add_argument('--log-stderr',
                            action='store_true',
                            help='log to stderr')
        parser.add_argument('--log-syslog',
                            action='store_true',
                            help='log to syslog')
        parser.add_argument('--syslog-socket',
                            default=default_syslog_socket(),
                            help='syslog server socket address. '
                                 'socket path, or host:port for UDP'
                            )
        parser.add_argument('--flush-interval',
                            type=float,
                            help='flush interval in seconds')
        parser.add_argument('--flush-stdout',
                            action='store_true',
                            help='flush to standard output')
        parser.add_argument('--flush-graphite',
                            help='comma separated graphite addresses to flush '
                                 'stats to, each in host[:port] format',
                            default=None)
        parser.add_argument('--flush-file',
                            help='pipe separated file paths to flush '
                                 'stats to',
                            default=None)
        parser.add_argument('--flush-file-csv',
                            help='pipe separated file paths to flush '
                                 'stats to in CSV format',
                            default=None)
        parser.add_argument('--collect-udp',
                            help='listen on UDP addresses to collect stats')
        parser.add_argument('--collect-tcp',
                            help='listen on TCP addresses to collect stats')
        parser.add_argument('--collector-threads',
                            help='number of threads started by each collector'
                                 ' (TCP collectors only)',
                            type=int
                            )
        parser.add_argument('--collector-threads-limit',
                            help='max number of threads running by each collector'
                                 ' (TCP collectors only)',
                            type=int
                            )

        return parser.parse_args(args)

    def _validate_configs(self, args):
        # type: (Dict[str, Any]) -> None
        none_negative_args = ('collector_threads_limit',)
        greater_than_one_args = ('collector_threads',)
        for key, value in args.items():
            if key in none_negative_args and value < 0:
                raise ValueError("The value for {} can not be negative".format(key))
            if key in greater_than_one_args and value < 1:
                raise ValueError("The value for {} can not be less than 1".format(key))
        if args['collector_threads_limit'] != 0 \
                and args['collector_threads_limit'] < args['collector_threads']:
            raise ValueError(
                "The value for collector_threads_limit can not be less than collector_threads")

    def _register_signal_handlers(self):
        # type: () -> None
        signal.signal(signal.SIGINT, self._handle_signal_int)
        signal.signal(signal.SIGTERM, self._handle_signal_term)
        # we can not set a handler for SIGHUP on Windows
        if hasattr(signal, 'SIGHUP'):
            try:
                signal.signal(signal.SIGHUP, self._handle_signal_hup)
                pass
            except ValueError:
                pass

    def _handle_signal_int(self, *args):
        # type: (*Any) -> None
        self._log("received SIGINT")
        self.shutdown()

    def _handle_signal_term(self, *args):
        # type: (*Any) -> None
        self._log("received SIGTERM")
        self.shutdown()

    def _handle_signal_hup(self, *args):
        # type: (*Any) -> None
        self._log("received SIGHUP")
        self.reload()

    def _create_logger(self):
        # type: () -> logging.Logger
        log_level_name = self._config.get('log_level', 'INFO')
        if log_level_name not in LOG_LEVEL_NAMES:
            raise ValueError("invalid log level " + log_level_name)
        logger = logging.Logger('navdoon')  # type: ignore
        logger.addHandler(logging.NullHandler())
        logger.setLevel(getattr(logging, log_level_name))
        if self._config.get('log_stderr'):
            logger.addHandler(logging.StreamHandler(sys.stderr))
        if self._config.get('log_file'):
            logger.addHandler(logging.FileHandler(self._config['log_file']))
        if self._config.get('log_syslog'):
            syslog_address = self._config['syslog_socket'].split(':')
            if len(syslog_address) < 2:
                syslog_address = syslog_address[0].strip()
            else:
                syslog_address = tuple([syslog_address[0].strip(),
                                        int(syslog_address[1])])
            logger.addHandler(logging.handlers.SysLogHandler(syslog_address))
        return logger

    def _close_logger(self):
        # type: () -> None
        if self.logger:
            handlers = []
            for handler in self.logger.handlers:
                if hasattr(handler, 'close') and callable(handler.close):
                    handler.close()
                handlers.append(handler)
            map(self.logger.removeHandler, handlers)
        self.logger = None

    @staticmethod
    def get_addresses_with_unique_ports(addresses):
        # type: (str) -> List[Tuple[str, int]]
        address_tuples = [tuple(address.strip().split(':')) for address in
                          addresses.split(',')]
        result = []  # type: List[Tuple[str, int]]
        ports = set()  # type: Set[int]
        for address in address_tuples:
            host = address[0]
            if len(address) > 1 and address[1]:
                port_str = address[1]
                port = int(port_str)
                if port < 1 or port > 65535:
                    raise ValueError(
                        "Port {} is out of range".format(port_str))
                if port in ports:
                    raise ValueError(
                        "Port {} is already specified before".format(port_str))
            else:
                port = DEFAULT_PORT
            result.append((host, port))
            ports.add(port)
        return result


def main(args=sys.argv[1:]):
    # type: (List[str]) -> int
    """Entry point, setup and start the application"""
    app = App(args)
    return app.run()


if __name__ == '__main__':
    sys.exit(main())



"""
navdoon.destination.file
------------------------
Define destinations to flush metrics to files
"""

from navdoon.destination.stream import Stream, CsvStream
from navdoon.pystdlib.typing import Any


class TextFile(Stream):
    """Destination to flush metrics to a file"""

    def __init__(self, name):
        # type: (str) -> None
        self._name = None  # type: str
        self.name = name
        file_handle = open(name, 'at')
        Stream.__init__(self, file_handle)

    def __del__(self):
        self.stream.close()

    @property
    def name(self):
        # type: () -> str
        return self._name

    @name.setter
    def name(self, name):
        # type: (str) -> None
        self._name = name

    def __eq__(self, other):
        # type: (Any) -> bool
        return self._name == other.name and self.pattern == other.pattern


class CsvFile(TextFile):
    """Destination to flush metrics to a CSV file"""

    def __init__(self, name):
        # type: (str) -> None
        TextFile.__init__(self, name)
        self.pattern = '"{name}","{value}","{timestamp}"'  # type: str
        self.append = "\r\n"  # type: str



"""
navdoon.processor
-----------------
Define queue processor, that will process the actualy Statsd requests
queued by the collectors.
"""

from time import time
from threading import Event, RLock, Thread
from navdoon.pystdlib.queue import Empty, Queue
from navdoon.utils.common import LoggerMixIn, DataSeries
from navdoon.pystdlib.typing import List, Any, Tuple, Dict, Set
from navdoon.destination import AbstractDestination
from statsdmetrics import (Counter, Gauge, GaugeDelta, Timer, parse_metric_from_request)
from statsdmetrics import Set as SetMetric


def validate_destinations(destinations):
    for destination in destinations:
        if not hasattr(destination,
                       'flush') or not callable(destination.flush):
            raise ValueError("Invalid destination for queue processor."
                             "Destination should have a flush() method")


def validate_queue(queue_):
    # type: (Any) -> None
    if not callable(getattr(queue_, 'get', None)):
        raise ValueError("Invalid queue for queue processor."
                         "queue should have a get() method")


class QueueProcessor(LoggerMixIn):
    """Process Statsd requests queued by the collectors"""

    def __init__(self, queue_):
        # type: (Queue) -> None
        validate_queue(queue_)
        super(QueueProcessor, self).__init__()
        self.log_signature = 'queue.processor '  # type: str
        self.stop_process_token = None  # type: str
        self._flush_interval = 1  # type: float
        self._queue = queue_  # type: Queue
        self._should_stop_processing = Event()  # type: Event
        self._processing = Event()  # type: Event
        self._shutdown = Event()  # type: Event
        self._processing_lock = RLock()  # type: RLock
        self._flush_lock = RLock()  # type: RLock
        self._shelf = StatsShelf()  # type: StatsShelf
        self._destinations = []  # type: List[AbstractDestination]
        self._flush_queues = []  # type: List[Queue]
        self._flush_threads = []  # type: List[Thread]
        self._flush_threads_initialized = Event()  # type: Event
        self._should_stop_flushing = Event()  # type: Event
        self._last_flush_timestamp = None  # type: float

    @property
    def queue(self):
        # type: () -> Queue
        return self._queue

    @queue.setter
    def queue(self, queue_):
        # type: (Queue) -> None
        validate_queue(queue_)
        if self.is_processing():
            raise Exception(
                "Can not change queue processor's queue while processing")
        self._queue = queue_

    @property
    def flush_interval(self):
        # type: () -> float
        return self._flush_interval

    @flush_interval.setter
    def flush_interval(self, interval):
        # type: (float) -> None
        interval_float = float(interval)
        if interval_float <= 0:
            raise ValueError(
                "Invalid flush interval. Interval should be a positive number")
        self._flush_interval = interval_float

    def set_destinations(self, destinations):
        # type: (List[AbstractDestination]) -> QueueProcessor
        validate_destinations(destinations)
        self._destinations = destinations
        return self

    def clear_destinations(self):
        # type: () -> QueueProcessor
        self._destinations = []
        return self

    def get_destinations(self):
        # type: () -> List[AbstractDestination]
        return self._destinations

    def init_destinations(self):
        # type: () -> None
        self._log_debug("initializing destination ...")
        self._flush_threads_initialized.set()
        self._stop_flush_threads()
        self._clear_flush_threads()

        self._log_debug("initializing {} destination threads ...".format(len(self._destinations)))
        for destination in self._destinations:
            queue_ = Queue()  # type: Queue
            flush_thread = Thread(
                target=self._flush_metrics_queue_to_destination,
                args=(queue_, destination))
            flush_thread.setDaemon(True)
            flush_thread.start()
            self._flush_queues.append(queue_)
            self._flush_threads.append(flush_thread)

        self._flush_threads_initialized.set()
        self._log_debug("initialized {} destination threads".format(len(self._flush_threads)))

    def destinations_initialized(self):
        # type: () -> bool
        return self._flush_threads_initialized.is_set()

    def is_processing(self):
        # type: () -> bool
        return self._processing.is_set()

    def wait_until_processing(self, timeout=None):
        # type: (float) -> None
        self._processing.wait(timeout)

    def process(self):
        # type: () -> None
        self._log_debug("waiting for process lock ...")
        with self._processing_lock:
            self._log_debug("process lock acquired")
            if self._last_flush_timestamp is None:
                self._last_flush_timestamp = time()
            self._log("processing the queue ...")

            if not self.destinations_initialized():
                self.init_destinations()

            queue_get = self._queue.get
            QueueEmptyError = Empty
            process = self._process_request
            should_stop = self._should_stop_processing.is_set
            log_debug = self._log_debug
            flush = self.flush
            flush_interval = self._flush_interval

            self._shutdown.clear()
            self._processing.set()

            try:
                while True:
                    if should_stop():
                        log_debug("instructed to shutdown. stopping processing ...")
                        break
                    try:
                        data = queue_get(timeout=1)
                        queue_has_data = True
                    except QueueEmptyError:
                        queue_has_data = False

                    if float(time() - self._last_flush_timestamp) >= flush_interval:
                        flush()

                    if queue_has_data:
                        if data == self.stop_process_token:
                            self._log("got stop process token in queue")
                            break
                        elif data:
                            process(data)
            finally:
                self._should_stop_processing.clear()
                self._processing.clear()
                self._stop_flush_threads()
                self._clear_flush_threads()
                self._flush_queues = []
                self._shutdown.set()
                self._log("stopped processing the queue")

    def flush(self):
        # type: () -> None
        self._log_debug("waiting for flush lock")
        with self._flush_lock:
            self._log_debug("flushing lock acquired")
            now = time()
            metrics = self._get_metrics_and_clear_shelf(now)
            for queue_ in self._flush_queues:
                queue_.put(metrics)
            self._last_flush_timestamp = now
            self._log("flushed {} metrics to {} queues".format(len(metrics), len(self._flush_queues)))

    def shutdown(self):
        # type: () -> None
        self._log("shutting down ...")
        self._should_stop_processing.set()
        self._stop_flush_threads()

    def wait_until_shutdown(self, timeout=None):
        # type: (float) -> None
        self._shutdown.wait(timeout)

    def _flush_metrics_queue_to_destination(self, queue_, destination):
        # type: (Queue, AbstractDestination) -> None
        should_stop = self._should_stop_flushing.is_set
        queue_get = queue_.get
        flush = destination.flush
        QueueEmptyError = Empty
        while not should_stop():
            try:
                flush(queue_get(timeout=1))
                self._log_debug("flushed metrics to destination {}".format(destination))
            except QueueEmptyError:
                pass
        self._log("stopped flushing metrics to destination {}".format(destination))

    def _process_request(self, request):
        # type: (str) -> None
        request = str(request)
        self._log_debug("processing metrics: {}".format(request))
        lines = [line.strip() for line in request.split("\n") if line.strip()]
        should_stop = self._should_stop_processing.is_set
        for line in lines:
            if should_stop():
                break
            try:
                metric = parse_metric_from_request(line)
            except ValueError as parse_error:
                self._log_error(
                    "failed to parse statsd metrics from '{}': {}".format(
                        line, parse_error))
                continue
            self._shelf.add(metric)

    def _get_metrics_and_clear_shelf(self, timestamp):
        # type: (float) -> List[Tuple[str, float, float]]
        shelf = self._shelf
        counters = shelf.counters()
        gauges = shelf.gauges()
        sets = shelf.sets()
        timers = shelf.timers()
        shelf.clear()

        metrics = []
        for name, value in counters.items():
            metrics.append((name, value, timestamp))

        for name, value in gauges.items():
            metrics.append((name, value, timestamp))

        for name, values in sets.items():
            metrics.append((name, len(values), timestamp))

        for name, timer_stats in timers.items():
            for statistic, value in timer_stats.items():
                metrics.append(
                    (
                        "{}.{}".format(name, statistic),
                        value,
                        timestamp
                    )
                )

        return metrics

    def _stop_flush_threads(self):
        # type: () -> QueueProcessor
        self._log_debug("flush threads should stop")
        self._should_stop_flushing.set()
        return self

    def _clear_flush_threads(self):
        # type: () -> QueueProcessor
        self._log_debug("clearing {} flush threads".format(len(self._flush_threads)))
        for thread in self._flush_threads:
            if thread:
                thread.join(5)
        self._flush_threads = []
        self._flush_queues = []
        self._should_stop_flushing.clear()
        self._flush_threads_initialized.clear()
        return self


class StatsShelf(object):
    """A container that will aggregate and accumulate metrics"""

    _metric_add_methods = {Counter.__name__: '_add_counter',
                           SetMetric.__name__: '_add_set',
                           Gauge.__name__: '_add_gauge',
                           GaugeDelta.__name__: '_add_gauge_delta',
                           Timer.__name__: '_add_timer'}  # type: Dict[str, str]

    def __init__(self):
        # type: () -> None
        self._lock = RLock()  # type: RLock
        self._counters = dict()  # type: Dict[str, float]
        self._timers = dict()  # type: Dict[str, List[float]]
        self._sets = dict()  # type: Dict[str, Set[Any]]
        self._gauges = dict()  # type: Dict[str, float]

    def add(self, metric):
        # type: (Any) -> None
        method_name = self._metric_add_methods.get(
            metric.__class__.__name__)
        if not method_name:
            raise ValueError(
                "Can not add metric to shelf. No method is defined to "
                "handle {}".format(metric.__class__.__name__))
        with self._lock:
            getattr(self, method_name)(metric)

    def counters(self):
        # type: () -> Dict[str, float]
        return self._counters.copy()

    def sets(self):
        # type: () -> Dict[str, Set[Any]]
        return self._sets.copy()

    def gauges(self):
        # type: () -> Dict[str, float]
        return self._gauges.copy()

    def timers_data(self):
        # type: () -> Dict[str, List[float]]
        return self._timers.copy()

    def timers(self):
        # type: () -> Dict[str, Dict[str, float]]
        result = dict()
        for name, timers in self._timers.items():
            series = DataSeries(timers)
            result[name] = dict(count=series.count(), min=series.min(), max=series.max(),
                                mean=series.mean(), median=series.median())
        return result

    def clear(self):
        # type: () -> None
        self._counters = dict()
        self._timers = dict()
        self._sets = dict()
        self._gauges = dict()

    def _add_counter(self, counter):
        # type: (Counter) -> None
        counters = self._counters
        name = counter.name
        if name not in counters:
            counters[name] = 0
        counters[name] += counter.count / counter.sample_rate

    def _add_set(self, metric):
        # type: (SetMetric) -> None
        self._sets.setdefault(metric.name, set()).add(metric.value)

    def _add_gauge(self, metric):
        # type: (Gauge) -> None
        self._gauges[metric.name] = metric.value

    def _add_gauge_delta(self, metric):
        # type: (GaugeDelta) -> None
        gauges = self._gauges
        name = metric.name
        if name not in gauges:
            gauges[name] = metric.delta
        else:
            gauges[name] += metric.delta

    def _add_timer(self, metric):
        # type: (Timer) -> None
        self._timers.setdefault(metric.name, list()).append(
            metric.milliseconds)



import os
import unittest
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from os import remove
from time import time
from tempfile import mkstemp
from navdoon.destination import Graphite, Stream, CsvStream, TextFile


class TestGraphite(unittest.TestCase):
    def test_create_request_from_metrics(self):
        metrics = [('users', 34, 123456), ('cpu', 78, 98765)]
        self.assertEqual(["users 34 123456", "cpu 78 98765"],
                         Graphite.create_request_from_metrics(metrics), )

        metrics = [('no.time', 34), ('is.fine', 78, time())]
        self.assertEqual(2, len(Graphite.create_request_from_metrics(metrics)))

    def test_equality_based_on_attrs(self):
        graphite1 = Graphite('example.org', 2003)
        graphite2 = Graphite('localhost', 2004)
        self.assertNotEqual(graphite1, graphite2)

        graphite1.host = 'localhost'
        graphite1.port = 2004
        self.assertEqual(graphite1, graphite2)


class TestStream(unittest.TestCase):
    def test_stream_property(self):
        output = StringIO()
        dest = Stream(output)
        self.assertEqual(dest.stream, output)

        another_dest = Stream(StringIO())
        self.assertNotEqual(another_dest.stream, output)
        another_dest.stream = output
        self.assertEqual(another_dest.stream, output)

    def test_stream_property_fail_on_invalid_object(self):
        self.assertRaises(ValueError, Stream, "not a file like object")

    def test_equality_based_on_stream(self):
        output1 = StringIO()
        output2 = StringIO()
        stream1 = Stream(output1)
        stream2 = Stream(output2)
        self.assertNotEqual(stream1, stream2)
        stream2.stream = output1
        self.assertEqual(stream1, stream2)

    def test_create_output_from_metrics(self):
        output = StringIO()
        dest = Stream(output)

        metrics = [('logins', 12, 456789), ('mem', 53, 98765)]
        self.assertEqual(["logins 12 456789", "mem 53 98765"],
                         dest.create_output_from_metrics(metrics), )

        metrics = [('no.time', 192), ('is.fine', 221, time())]
        self.assertEqual(2, len(dest.create_output_from_metrics(metrics)))

    def test_flush(self):
        output = StringIO()
        dest = Stream(output)

        metrics = [('logins', 12, 456789), ('mem', 53, 98765)]
        dest.flush(metrics)
        self.assertEqual("logins 12 456789\nmem 53 98765\n", output.getvalue())

    def test_flush_with_pattern(self):
        output = StringIO()
        dest = Stream(output)
        dest.pattern = '"{name}"={value}@{timestamp}'

        metrics = [('logins', 12, 456789), ('mem', 53, 98765)]
        dest.flush(metrics)
        self.assertEqual('"logins"=12@456789\n"mem"=53@98765\n',
                         output.getvalue())

    def test_flush_with_append(self):
        output = StringIO()
        dest = Stream(output)
        dest.append = "+++"

        metrics = [('users', 800, 5678), ('cpu', 99, 1234)]
        dest.flush(metrics)
        self.assertEqual("users 800 5678+++cpu 99 1234+++", output.getvalue())

    def test_flush_with_partial_pattern_and_append(self):
        output = StringIO()
        dest = Stream(output)
        dest.pattern = "({name}:'{value}'"
        dest.append = ")"

        metrics = [('users', 800, 5678), ('cpu', 99, 1234)]
        dest.flush(metrics)
        self.assertEqual("(users:'800')(cpu:'99')", output.getvalue())


class TestCsvStream(unittest.TestCase):
    def test_create_output_from_metrics(self):
        output = StringIO()
        dest = CsvStream(output)

        metrics = [('events', 8, 12345), ('mem', 34, 98765)]
        self.assertEqual(['"events","8","12345"', '"mem","34","98765"'],
                         dest.create_output_from_metrics(metrics), )

        metrics = [('no.time', 192), ('is.fine', 221, time())]
        self.assertEqual(2, len(dest.create_output_from_metrics(metrics)))

    def test_flush(self):
        output = StringIO()
        dest = CsvStream(output)
        metrics = [('logins', 12, 456789), ('mem', 53, 98765)]
        dest.flush(metrics)
        self.assertEqual('"logins","12","456789"\r\n"mem","53","98765"\r\n', output.getvalue())


class TestTextFile(unittest.TestCase):
    def setUp(self):
        _, self.temp_file_name = mkstemp()
        os.remove(self.temp_file_name)

    def tearDown(self):
        if os.path.exists(self.temp_file_name):
            remove(self.temp_file_name)

    def test_flush(self):
        dest = TextFile(self.temp_file_name)
        metrics = [('logins', 12, 456789), ('mem', 53, 98765)]
        dest.flush(metrics)
        self.assertTrue(os.path.exists(self.temp_file_name))
        self.assertFileHasLine('logins 12 456789', self.temp_file_name)
        self.assertFileHasLine('mem 53 98765', self.temp_file_name)

    def assertFileHasLine(self, expected, filename):
        with open(filename) as file_:
            lines = [line.rstrip() for line in file_.readlines()]
            self.assertIn(expected, lines)


import os
import socket
import sys
import logging
import logging.handlers
import unittest
from os import path
from tempfile import mkstemp

from navdoon.app import App, parse_config_file, default_syslog_socket
from navdoon.collector import SocketServer
from navdoon.destination import Stdout, Graphite, TextFile, CsvFile
from navdoon.processor import QueueProcessor

test_config_file_path = path.join(
    path.join(
        path.dirname(__file__), 'fixtures'), 'test_config.ini')


class TestParseConfigFile(unittest.TestCase):
    def test_parse_config_file(self):
        with open(test_config_file_path) as test_config_file:
            config = parse_config_file(test_config_file)
        self.assertEqual('WARN', config['log_level'])


class TestDefaultSyslogSocket(unittest.TestCase):
    def test_default_syslog_socket(self):
        syslog_socket = default_syslog_socket()
        self.assertIsInstance(syslog_socket, str)
        self.assertNotEqual(syslog_socket, '')
        if not hasattr(self, 'assertRegex'):
            self.assertRegex = self.assertRegexpMatches
        self.assertRegex(syslog_socket, '[/:]')


class TestApp(unittest.TestCase):
    def setUp(self):
        self.config_filename = test_config_file_path

    def test_get_args(self):
        app = App(())
        self.assertEqual((), app.get_args())

        args = ('--log-level', 'DEBUG')
        app = App(args)
        self.assertEqual(args, app.get_args())

    def test_get_config(self):
        app = App(())
        conf = app.get_config()
        self.assertFalse(conf['log_syslog'])

        app = App(('--log-syslog', '-c', self.config_filename))
        conf = app.get_config()
        self.assertTrue(conf['log_syslog'])
        self.assertEqual('WARN', conf['log_level'])

    def test_args_overwrite_config(self):
        app = App(('--log-level', 'ERROR', '-c', self.config_filename))
        conf = app.get_config()
        self.assertEqual('ERROR', conf['log_level'])

    def test_validate_configs(self):
        self.assertRaises(ValueError, App, ('--collector-threads', '0'))
        self.assertRaises(ValueError, App, ('--collector-threads-limit', '-1'))
        self.assertRaises(
            ValueError, App,
            ('--collector-threads', '2', '--collector-threads-limit', '1')
        )

    def test_defaults_for_optional_args(self):
        app = App(())
        conf = app.get_config()
        self.assertEqual('INFO', conf['log_level'])
        self.assertFalse(conf['log_stderr'])
        self.assertEqual(None, conf['log_file'])
        self.assertFalse(conf['log_syslog'])
        self.assertLess(0, conf['flush_interval'])
        self.assertFalse(conf['flush_stdout'])
        self.assertEqual('', conf['flush_graphite'])
        self.assertEqual('', conf['flush_file'])
        self.assertEqual('', conf['collect_udp'])
        self.assertEqual('', conf['collect_tcp'])

    def test_default_options_cover_missing_configs(self):
        app = App(('--log-level', 'FATAL', '-c', self.config_filename))
        conf = app.get_config()
        self.assertEqual('FATAL', conf['log_level'])
        self.assertTrue(conf['log_stderr'])
        self.assertEqual(None, conf['log_file'])
        self.assertFalse(conf['log_syslog'])
        self.assertLess(0, conf['flush_interval'])
        self.assertTrue(conf['flush_stdout'])
        self.assertEqual('', conf['flush_graphite'])
        self.assertEqual('', conf['collect_udp'])
        self.assertEqual('', conf['collect_tcp'])

    def test_get_logger(self):
        app = App(('--config', self.config_filename, '--log-syslog'))
        logger = app.get_logger()
        self.assertEqual(logging.WARN, logger.getEffectiveLevel())
        self.assertEqual(3, len(logger.handlers))
        null_handler, stderr_handler, syslog_handler = logger.handlers
        self.assertIsInstance(stderr_handler, logging.StreamHandler)
        self.assertEqual(stderr_handler.stream, sys.stderr)
        self.assertIsInstance(syslog_handler, logging.handlers.SysLogHandler)

    def test_create_destinations(self):
        app = App(['--config', self.config_filename])
        destinations = app.create_destinations()
        self.assertEqual(1, len(destinations))
        self.assertEqual([Stdout()], destinations)

        app_flush_graphite = App(['--flush-graphite', 'localhost'])
        destinations = app_flush_graphite.create_destinations()
        self.assertEqual(1, len(destinations))
        self.assertEqual([Graphite('localhost', 2003)], destinations)

        app_multi_flush = App(
            ['--config', self.config_filename, '--flush-graphite',
             'example.org:2006,localhost'])
        destinations = app_multi_flush.create_destinations()
        self.assertEqual(3, len(destinations))
        expected = [
            Stdout(), Graphite('example.org', 2006),
            Graphite ('localhost', 2003)
        ]
        self.assertEqual(expected, destinations)

    def test_create_file_destinations(self):
        try:
            _, temp_file_name = mkstemp()
            other_temp_file_name = temp_file_name + '_2'
            app = App(['--flush-file', '{}|{}'.format(temp_file_name, other_temp_file_name)])
            destinations = app.create_destinations()
            self.assertEqual(2, len(destinations))
            self.assertEqual([TextFile(temp_file_name), TextFile(other_temp_file_name)], destinations)
        finally:
            if os.path.exists(temp_file_name):
                os.remove(temp_file_name)
            if os.path.exists(other_temp_file_name):
                os.remove(other_temp_file_name)

    def test_create_csv_file_destinations(self):
        try:
            _, temp_file_name = mkstemp()
            other_temp_file_name = temp_file_name + '_2'
            app = App(['--flush-file-csv', '{}|{}'.format(temp_file_name, other_temp_file_name)])
            destinations = app.create_destinations()
            self.assertEqual(2, len(destinations))
            self.assertEqual([CsvFile(temp_file_name), CsvFile(other_temp_file_name)], destinations)
        finally:
            if os.path.exists(temp_file_name):
                os.remove(temp_file_name)
            if os.path.exists(other_temp_file_name):
                os.remove(other_temp_file_name)

    def test_get_addresses_with_unique_ports(self):
        app = App([])
        self.assertEqual(
            [("localhost", 8127)],
            app.get_addresses_with_unique_ports("localhost:8127")
        )
        self.assertEqual(
            [("example.org", 8125), ("localhost", 8126), ("127.0.0.1", 8127)],
            app.get_addresses_with_unique_ports("example.org,localhost:8126,127.0.0.1:8127")
        )
        self.assertEqual(
            [("", 12345), ("example.org", 8125), ("", 6789)],
            app.get_addresses_with_unique_ports(":12345,example.org,:6789")
        )
        self.assertRaises(ValueError, app.get_addresses_with_unique_ports, "localhost:8125,127.0.0.1:8125")
        self.assertRaises(ValueError, app.get_addresses_with_unique_ports, "localhost:81250000")
        self.assertRaises(ValueError, app.get_addresses_with_unique_ports, "localhost:0")
        self.assertRaises(ValueError, app.get_addresses_with_unique_ports, "localhost:-23")
        self.assertRaises(ValueError, app.get_addresses_with_unique_ports, "localhost,:8125")

    def test_create_collectors_creates_udp_server_by_default(self):
        app = App([])
        collectors = app.create_collectors()
        self.assertEqual(len(collectors), 1)
        collector = collectors.pop()
        self.assertIsInstance(collector, SocketServer)
        self.assertEqual(
            ("127.0.0.1", 8125, socket.SOCK_DGRAM),
            (collector.host, collector.port, collector.socket_type)
        )

    def test_create_udp_collectors(self):
        app = App(['--collect-udp', ':8127,example.com,127.0.0.1:8126'])
        collectors = app.create_collectors()
        self.assertEqual(len(collectors), 3)
        for collector in collectors:
            self.assertIsInstance(collector, SocketServer)
        self.assertEqual(
            ("", 8127, socket.SOCK_DGRAM),
            (collectors[0].host, collectors[0].port, collectors[0].socket_type)
        )
        self.assertEqual(
            ("example.com", 8125, socket.SOCK_DGRAM),
            (collectors[1].host, collectors[1].port, collectors[1].socket_type)
        )
        self.assertEqual(
            ("127.0.0.1", 8126, socket.SOCK_DGRAM),
            (collectors[2].host, collectors[2].port, collectors[2].socket_type)
        )

    def test_create_server(self):
        app = App(['--config', self.config_filename, '--flush-interval', '17'])
        logger = app.get_logger()
        server = app.create_server()
        queue_processor = server.queue_processor
        self.assertEqual(server.logger, logger)
        self.assertIsInstance(queue_processor, QueueProcessor)
        self.assertEqual(queue_processor.logger, logger)
        self.assertEqual(queue_processor.flush_interval, 17)

    def test_create_tcp_collectors(self):
        app = App(['--collect-tcp', ':8127,example.org,127.0.0.1:8126',
                   '--collector-threads', '8', '--collector-threads-limit', '32'])
        collectors = app.create_collectors()
        self.assertEqual(len(collectors), 3)
        for collector in collectors:
            self.assertIsInstance(collector, SocketServer)
            self.assertEqual(collector.num_worker_threads, 8)
            self.assertEqual(collector.worker_threads_limit, 32)
        self.assertEqual(
            ("", 8127, socket.SOCK_STREAM),
            (collectors[0].host, collectors[0].port, collectors[0].socket_type)
        )
        self.assertEqual(
            ("example.org", 8125, socket.SOCK_STREAM),
            (collectors[1].host, collectors[1].port, collectors[1].socket_type)
        )
        self.assertEqual(
            ("127.0.0.1", 8126, socket.SOCK_STREAM),
            (collectors[2].host, collectors[2].port, collectors[2].socket_type)
        )



"""
navdoon.pystdlib.typing
----------------------
Type hints from Python standard library. This is to remove requirements for Typing module to be available
on production.
"""
from __future__ import absolute_import

try:
    from typing import List, Dict, Tuple, AnyStr, IO, Any, Sequence, Set, Union, Callable, Optional, Mapping
except ImportError:
    List, Dict, Tuple, AnyStr, Any, IO, Sequence = None, None, None, None, None, None, None  # type: ignore
    Mapping, Set, Union, Callable, Optional = None, None, None, None, None  # type: ignore



#!/usr/bin/env python
"""
navdoon
"""

from __future__ import print_function

import os
from setuptools import setup, find_packages

try:
    import distutilazy.test, distutilazy.clean  # type: ignore
except ImportError:
    distutilazy = None  # type: ignore


from navdoon import (__title__, __summary__, __version__, __author__,
                     __license__)

__all__ = ['setup_params', 'classifiers', 'long_description']  # type: ignore

classifiers = [
    "Development Status :: 4 - Beta", "Intended Audience :: Developers",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: System :: Networking :: Monitoring"
]  # type: ignore

long_description = __summary__  # type: str
with open(os.path.join(os.path.dirname(__file__), "README.rst")) as fh:
    long_description = fh.read()

setup_params = dict(name=__title__,
                    packages=find_packages(),
                    version=__version__,
                    description=__summary__,
                    long_description=long_description,
                    author=__author__,
                    url="https://github.com/farzadghanei/navdoon",
                    license=__license__,
                    classifiers=classifiers,
                    install_requires=["statsdmetrics>=0.3"],
                    keywords="statsd monitoring",
                    test_suite="tests",
                    zip_safe=False,
                    extras_require=dict(
                        dev=["distutilazy>=0.4.2", "typing>=3.5.0.1", "coverage"]
                    ),
                    entry_points=dict(
                        console_scripts=["navdoon=navdoon.app:main", ]
                    ),
                    scripts=['bin/navdoon']
                    )  # type: ignore

if distutilazy:
    setup_params["cmdclass"] = dict(test=distutilazy.test.run_tests,
                                    clean_pyc=distutilazy.clean.clean_pyc,
                                    clean=distutilazy.clean.clean_all)

if __name__ == "__main__":
    setup(**setup_params)


"""
navdoon.utils.system
--------------------
System utilities and mixin classes
"""

import platform
from time import time
from multiprocessing import cpu_count
from threading import Thread, RLock, Event
from navdoon.pystdlib.queue import Queue, Empty
from navdoon.pystdlib.typing import Dict, Callable, List, Any, Sequence
from navdoon.utils.common import LoggerMixIn

PLATFORM_NAME = platform.system().strip().lower()


def available_cpus():
    # type: () -> int
    try:
        cpus = cpu_count()
    except Exception:
        cpus = 1
    return cpus


def os_syslog_socket():
    # type: () -> str
    syslog_addresses = dict(
        linux="/dev/log",
        darwin="/var/run/syslog",
        freebsd="/var/run/log"
    )
    return syslog_addresses.get(PLATFORM_NAME, None)


class WorkerThread(Thread):
    """A thread to keep consuming tasks from a task queue and store
    the results in a dictionary of task_ids => results
    """

    def __init__(self, queue, stop_event, results):
        # type: (Queue, Event, Dict[int, Any]) -> None
        self.queue = queue  # type: Queue
        self.stop_event = stop_event  # type: Event
        self.results = results  # type: Dict[int, Any]
        Thread.__init__(self)

    def _consume_queue(self):
        # type: () -> None
        should_stop = self.stop_event.is_set
        while not should_stop():
            try:
                self._run_task_from_queue(timeout=1)
            except Empty:
                pass

    def _run_task_from_queue(self, timeout=None):
        # type: (float) -> None
        (task_id, func, args, kwargs) = self.queue.get(True, timeout)
        result = func(*args, **kwargs)
        self.results[task_id] = result
        self.queue.task_done()

    def run(self):
        # type: () -> None
        self._consume_queue()


class TemporaryWorkerThread(WorkerThread):
    """A worker thread that only consumes the queue as long as the queue is not
    empty.
    """

    def _consume_queue(self):
        # type: () -> None
        should_stop = self.stop_event.is_set
        while not should_stop():
            try:
                self._run_task_from_queue(1)
            except Empty:
                break


class ThreadPool(LoggerMixIn):
    def __init__(self, size):
        # type: (int) -> None
        LoggerMixIn.__init__(self)
        self._size = int(size)  # type: int
        self._threads = []  # type: List[Thread]
        self._queue = Queue()  # type: Queue
        self._queue_lock = RLock()  # type: RLock
        self._task_counter = 0  # type: int
        self._task_results = dict()  # type: Dict[int, Any]
        self._stop_event = Event()  # type: Event
        self.log_signature = "threadpool "  # type: str

    def __del__(self):
        # type: () -> None
        if not self._stop_event.is_set():
            self.stop()

    @property
    def size(self):
        # type: () -> int
        return self._size

    @property
    def threads(self):
        # type: () -> List[Thread]
        return self._threads

    def initialize(self):
        # type: () -> ThreadPool
        self._queue = Queue()
        self._create_worker_threads()
        self._start_worker_threads()
        return self

    def do(self, func, *args, **kwargs):
        # type: (Callable, *Any, **Any) -> int
        if self._stop_event.is_set():
            raise Exception("Task thread pool has stopped")
        with self._queue_lock:
            task_id = self._task_counter
            self._task_counter += 1
            self._handle_task(task_id, func, args, kwargs)
        return task_id

    def is_done(self):
        # type: () -> bool
        with self._queue_lock:
            is_done = self._queue.empty()
        return is_done

    def wait_until_done(self):
        # type: () -> ThreadPool
        self._queue.join()
        return self

    def stop(self, wait=True, timeout=None):
        # type: (bool, float) -> None
        self._stop_event.set()
        if wait:
            num_threads = len(self._threads)
            self._log_debug(
                "joining {} worker threads ...".format(num_threads))
            start_time = time()
            counter = 0
            for thread in self._threads:
                counter += 1
                self._log_debug("joining thread {} ...".format(counter))
                thread.join(timeout)
                if timeout is None:
                    continue
                elif time() - start_time > timeout:
                    raise Exception("Stopping thread pool timedout")
            self._log_debug("joined worker {} threads".format(num_threads))
            self._threads = []

    def get_result(self, task_id):
        # type: (int) -> Any
        if task_id not in self._task_results:
            raise ValueError(
                "No results found for task id '{}'".format(task_id))
        return self._task_results[task_id]

    def _handle_task(self, task_id, func, args, kwargs):
        # type: (int, Callable, Sequence[Any], Dict[Any, Any]) -> None
        self._queue.put((task_id, func, args, kwargs))

    def _create_worker_threads(self):
        # type: () -> None
        for i in range(self._size):
            worker = WorkerThread(self._queue, self._stop_event, self._task_results)
            self._threads.append(worker)

    def _start_worker_threads(self):
        # type: () -> ThreadPool
        for thread in self._threads:
            thread.start()
        self._stop_event.clear()
        return self


class ExpandableThreadPool(ThreadPool):
    def __init__(self, size, workers_limit=0):
        # type: (int, int) -> None
        ThreadPool.__init__(self, size)
        self._spawn_worker_threshold = 0.5  # type: float
        self._max_workers_count = 0  # type: int
        self._workers_limit = workers_limit  # type: int

    @property
    def workers_limit(self):
        # type: () -> int
        return self._workers_limit

    @workers_limit.setter
    def workers_limit(self, limit):
        # type: (int) -> None
        limit = int(limit)
        if limit < 0:
            raise ValueError("Thread pool workers limit can't be negative")
        self._workers_limit = limit

    @property
    def max_workers_count(self):
        # type: () -> int
        return self._max_workers_count

    @property
    def spawn_workers_threshold(self):
        # type: () -> float
        return self._spawn_worker_threshold

    @spawn_workers_threshold.setter
    def spawn_workers_threshold(self, threshold):
        # type: (float) -> None
        threshold = float(threshold)
        if threshold < 0:
            raise ValueError("Thread pool spawn workers threshold can't be negative")
        self._spawn_worker_threshold = threshold

    def _start_worker_threads(self):
        # type: () -> ExpandableThreadPool
        ThreadPool._start_worker_threads(self)
        self._max_workers_count = len(self._threads)
        return self

    def _handle_task(self, task_id, func, args, kwargs):
        # type: (int, Callable, Sequence[Any], Dict[Any, Any]) -> None
        ThreadPool._handle_task(self, task_id, func, args, kwargs)
        if not self._stop_event.is_set() and self._can_spawn_temp_worker():
            self._spawn_temp_worker()

    def _can_spawn_temp_worker(self):
        # type: () -> bool
        return self._queue.qsize() > (self._spawn_worker_threshold * self._size) \
               and (self._workers_limit == 0 or len(self._threads) < self._workers_limit)

    def _spawn_temp_worker(self):
        # type: () -> None
        thread = TemporaryWorkerThread(self._queue, self._stop_event, self._task_results)
        self._threads.append(thread)
        thread.start()
        self._max_workers_count += 1



import unittest
from time import sleep
from threading import Thread
import navdoon.utils.system
from navdoon.utils.system import ThreadPool, ExpandableThreadPool


def mock_cpu_count(count):
    def _cpu_count():
        return count

    return _cpu_count


def not_implemented(*args):
    raise NotImplementedError


class TestFunctions(unittest.TestCase):
    def test_available_cpus_returns_number_of_cpus(self):
        navdoon.utils.system.cpu_count = mock_cpu_count(3)
        self.assertEqual(3, navdoon.utils.system.available_cpus())

    def test_available_cpus_returns_minimum_count_on_errors(self):
        navdoon.utils.system.cpu_count = not_implemented
        self.assertEqual(1, navdoon.utils.system.available_cpus())


class TestThreadPool(unittest.TestCase):
    threadPoolClass = ThreadPool

    def test_init(self):
        pool = self.__class__.threadPoolClass(4)
        self.assertEqual(pool.size, 4)

    def test_initialize(self):
        pool = self.__class__.threadPoolClass(4)
        pool.initialize()
        self.assertEqual(len(pool.threads), 4)

    def test_do_tasks(self):
        pool = self.__class__.threadPoolClass(4)
        pool.initialize()
        executed = []

        def some_task():
            return executed.append(True)

        for i in range(10):
            pool.do(some_task)

        pool.wait_until_done()
        pool.stop()
        self.assertEqual(executed, [True] * 10)

    def test_check_task_results(self):
        pool = self.__class__.threadPoolClass(5)
        pool.initialize()

        def change_text(text, number):
            return "{}. {}".format(number, text.upper())

        words = ["this", "will be", "uppercased"]
        expected_results = ["1. THIS", "2. WILL BE", "3. UPPERCASED"]
        counter = 1
        task_ids = []

        for word in words:
            task_ids.append(pool.do(change_text, word, counter))
            counter += 1

        pool.wait_until_done()
        pool.stop()
        results = []

        for id in task_ids:
            results.append(pool.get_result(id))

        self.assertEqual(results, expected_results)

    def test_check_task_results_fails_on_invalid_task_id(self):
        pool = self.__class__.threadPoolClass(2)
        pool.initialize()
        pool.stop()
        self.assertRaises(ValueError, pool.get_result, 1003)


class TestExpandableThreadPool(TestThreadPool):
    threadPoolClass = ExpandableThreadPool

    def test_do_tasks_spawns_new_threads_when_workers_are_not_enough(self):
        pool = self.__class__.threadPoolClass(1)
        pool.spawn_workers_threshold = 0
        pool.workers_limit = 4
        pool.initialize()
        executed = []

        def long_running_task():
            sleep(0.2)
            return executed.append(True)

        def populate_queue(count):
            for i in range(count):
                pool.do(long_running_task)

        populate_thread = Thread(target=populate_queue, args=(10,))
        populate_thread.start()
        populate_thread.join()
        pool.wait_until_done()
        pool.stop()

        self.assertEqual(len(executed), 10)
        self.assertGreater(pool.max_workers_count, 1)
        self.assertLessEqual(pool.max_workers_count, 4)
